# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Image tests."""
import io
import unittest
from unittest import mock

import langfun.core as lf
from langfun.core.modalities import image as image_lib
from langfun.core.modalities import mime as mime_lib
import PIL.Image as pil_image
import pyglove as pg


image_content = (
    b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x18\x00\x00\x00\x18\x04'
    b'\x03\x00\x00\x00\x12Y \xcb\x00\x00\x00\x18PLTE\x00\x00'
    b'\x00fff_chaag_cg_ch^ci_ciC\xedb\x94\x00\x00\x00\x08tRNS'
    b'\x00\n\x9f*\xd4\xff_\xf4\xe4\x8b\xf3a\x00\x00\x00>IDATx'
    b'\x01c \x05\x08)"\xd8\xcc\xae!\x06pNz\x88k\x19\\Q\xa8"\x10'
    b'\xc1\x14\x95\x01%\xc1\n\xa143Ta\xa8"D-\x84\x03QM\x98\xc3'
    b'\x1a\x1a\x1a@5\x0e\x04\xa0q\x88\x05\x00\x07\xf8\x18\xf9'
    b'\xdao\xd0|\x00\x00\x00\x00IEND\xaeB`\x82'
)


def mock_request(*args, **kwargs):
  del args, kwargs
  return pg.Dict(content=image_content)


class ImageTest(unittest.TestCase):

  def test_from_bytes(self):
    image = image_lib.Image.from_bytes(image_content)
    self.assertEqual(image.image_format, 'png')
    self.assertIn('data:image/png;base64,', image._raw_html())
    self.assertEqual(image.to_bytes(), image_content)
    with self.assertRaisesRegex(
        lf.ModalityError, '.* cannot be converted to text'
    ):
      image.to_text()

  def test_from_bytes_invalid(self):
    image = image_lib.Image.from_bytes(b'bad')
    with self.assertRaisesRegex(ValueError, 'Expected MIME type'):
      _ = image.image_format

  def test_from_bytes_base_cls(self):
    self.assertIsInstance(
        mime_lib.Mime.from_bytes(image_content), image_lib.Image
    )

  def test_from_uri(self):
    image = image_lib.Image.from_uri('http://mock/web/a.png')
    with mock.patch('requests.get') as mock_requests_get:
      mock_requests_get.side_effect = mock_request
      self.assertEqual(image.image_format, 'png')
      self.assertEqual(
          image._raw_html(),
          '<img src="http://mock/web/a.png">'
      )
      self.assertEqual(image.to_bytes(), image_content)

  def test_from_uri_base_cls(self):
    with mock.patch('requests.get') as mock_requests_get:
      mock_requests_get.side_effect = mock_request
      image = mime_lib.Mime.from_uri('http://mock/web/a.png')
      self.assertIsInstance(image, image_lib.Image)
      self.assertEqual(image.image_format, 'png')
      self.assertEqual(
          image._raw_html(),
          '<img src="http://mock/web/a.png">'
      )
      self.assertEqual(image.to_bytes(), image_content)

  def test_image_size(self):
    image = image_lib.Image.from_uri('http://mock/web/a.png')
    with mock.patch('requests.get') as mock_requests_get:
      mock_requests_get.side_effect = mock_request
      self.assertEqual(image.size, (24, 24))

  def test_to_pil_image(self):
    image = image_lib.Image.from_uri('http://mock/web/a.png')
    with mock.patch('requests.get') as mock_requests_get:
      mock_requests_get.side_effect = mock_request
      self.assertIsInstance(image.to_pil_image(), pil_image.Image)

  def test_from_pil_image(self):
    image = pil_image.open(io.BytesIO(image_content))
    self.assertIsInstance(
        image_lib.Image.from_pil_image(image), image_lib.Image
    )

  def test_from_pil_image_os_error(self):
    img = pil_image.open(io.BytesIO(image_content))
    with mock.patch.object(img, 'save') as mock_save:
      mock_save.side_effect = [OSError, None]
      with mock.patch('os.chdir') as mock_chdir:
        with mock.patch('os.getcwd') as mock_getcwd:
          mock_getcwd.return_value = '/curr/dir'
          image = image_lib.Image.from_pil_image(img)
          self.assertIsInstance(image, image_lib.Image)
          self.assertEqual(mock_save.call_count, 2)
          mock_save.assert_has_calls([
              mock.call(mock.ANY, format='PNG'),
              mock.call(mock.ANY, format='PNG'),
          ])
          mock_chdir.assert_has_calls([
              mock.call('/tmp'),
              mock.call('/curr/dir'),
          ])

  def test_gif_is_compatible(self):
    # Create a simple 1x1 GIF image using PIL
    buf = io.BytesIO()
    img = pil_image.new('P', (1, 1))
    img.save(buf, format='GIF')
    gif_bytes = buf.getvalue()

    gif_image = image_lib.Image.from_bytes(gif_bytes)
    self.assertEqual(gif_image.mime_type, 'image/gif')

    # GIF should be compatible if PNG is in supported types
    self.assertTrue(gif_image._is_compatible(['image/png']))
    self.assertTrue(gif_image._is_compatible(['image/jpeg', 'image/webp']))
    self.assertTrue(gif_image._is_compatible(['image/png', 'image/jpeg']))

    # GIF should not be compatible if only unsupported types
    self.assertFalse(gif_image._is_compatible(['video/mp4']))
    self.assertFalse(gif_image._is_compatible(['application/pdf']))

  def test_gif_make_compatible(self):
    # Create a simple 1x1 GIF image using PIL
    buf = io.BytesIO()
    img = pil_image.new('P', (1, 1))
    img.save(buf, format='GIF')
    gif_bytes = buf.getvalue()

    gif_image = image_lib.Image.from_bytes(gif_bytes)
    self.assertEqual(gif_image.mime_type, 'image/gif')

    # Test 1: Convert to PNG (first priority when available)
    converted = gif_image.make_compatible(['image/png', 'image/jpeg'])
    self.assertEqual(converted.mime_type, 'image/png')
    self.assertIsInstance(converted, image_lib.Image)

    # Test 2: Convert to JPEG when PNG not supported
    converted = gif_image.make_compatible(['image/jpeg', 'image/webp'])
    self.assertEqual(converted.mime_type, 'image/jpeg')

    # Test 3: Convert to WEBP when PNG and JPEG not supported
    converted = gif_image.make_compatible(['image/webp'])
    self.assertEqual(converted.mime_type, 'image/webp')

    # Test 4: Should raise error when no compatible format
    with self.assertRaises(lf.ModalityError):
      gif_image.make_compatible(['video/mp4'])

  def test_is_compatible_direct_match(self):
    image = image_lib.Image.from_bytes(image_content)  # image/png
    self.assertTrue(image._is_compatible(['image/png', 'image/jpeg']))
    self.assertTrue(image._is_compatible(['image/png']))
    self.assertFalse(image._is_compatible(['image/jpeg']))

  def test_make_compatible_no_conversion(self):
    image = image_lib.Image.from_bytes(image_content)  # image/png
    converted_image = image.make_compatible(['image/png', 'image/jpeg'])
    self.assertIs(image, converted_image)

  def test_convert_to_format_jpeg_transparency(self):
    # Create a simple RGBA PNG image
    buf = io.BytesIO()
    img = pil_image.new('RGBA', (1, 1), (255, 0, 0, 128))
    img.save(buf, format='PNG')
    rgba_png_bytes = buf.getvalue()

    rgba_image = image_lib.Image.from_bytes(rgba_png_bytes)
    self.assertEqual(rgba_image.mime_type, 'image/png')

    # Convert to JPEG, should trigger transparency handling
    converted_image = rgba_image._convert_to_format('JPEG')
    self.assertEqual(converted_image.mime_type, 'image/jpeg')
    pil_img = converted_image.to_pil_image()
    self.assertEqual(pil_img.mode, 'RGB')

  def test_convert_to_format_os_error(self):
    image = image_lib.Image.from_bytes(image_content)
    mock_pil_image = mock.MagicMock()
    mock_save = mock_pil_image.save
    mock_save.side_effect = [OSError, None]

    with mock.patch.object(
        image, 'to_pil_image', return_value=mock_pil_image
    ), mock.patch('os.chdir') as mock_chdir, mock.patch(
        'os.getcwd'
    ) as mock_getcwd:
      mock_getcwd.return_value = '/curr/dir'
      converted_image = image._convert_to_format('PNG')
      self.assertIsInstance(converted_image, image_lib.Image)
      self.assertEqual(mock_save.call_count, 2)
      mock_save.assert_has_calls([
          mock.call(mock.ANY, format='PNG'),
          mock.call(mock.ANY, format='PNG'),
      ])
      mock_chdir.assert_has_calls([
          mock.call('/tmp'),
          mock.call('/curr/dir'),
      ])


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Environment event logger."""

from typing import Annotated
from langfun.env import interface
import pyglove as pg


_METRIC_NAMESPACE = '/langfun/env'


class MetricWriter(pg.Object, interface.EventHandler):
  """Event handler for streamz metrics."""

  app: Annotated[
      str,
      'Application name that will be used as a metric parameter.'
  ] = ''

  def _get_counter(
      self,
      name: str,
      description: str,
      parameters: dict[str, type[str]] | None = None,
  ) -> pg.monitoring.Counter:
    return self._metric_collection.get_counter(
        name=name,
        description=description,
        parameters=parameters,
    )

  def _get_scalar(
      self,
      name: str,
      description: str,
      parameters: dict[str, type[str]] | None = None
  ) -> pg.monitoring.Metric:
    return self._metric_collection.get_scalar(
        name=name,
        description=description,
        parameters=parameters
    )

  def _get_distribution(
      self,
      name: str,
      description: str,
      parameters: dict[str, type[str]] | None = None
  ) -> pg.monitoring.Metric:
    return self._metric_collection.get_distribution(
        name=name,
        description=description,
        parameters=parameters
    )

  def _error_tag(self, error: BaseException | None) -> str:
    if error is None:
      return 'Success'
    return pg.utils.ErrorInfo.from_exception(error).tag

  def _initialize_metrics(self) -> None:
    """Initializes metrics."""

    self._metric_collection = pg.monitoring.metric_collection(_METRIC_NAMESPACE)

    #
    # Environment metrics.
    #

    self._environment_housekeep_duration_ms = self._get_distribution(
        'environment_housekeep_duration_ms',
        description='Environment housekeeping duration in milliseconds',
        parameters={
            'app': str,
            'environment_id': str,
            'error': str,
        }
    )

    #
    # Sandbox metrics.
    #

    # Sandbox counters.
    self._sandbox_start = self._get_counter(
        'sandbox_start',
        description='Sandbox start counter',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'error': str,
        }
    )
    self._sandbox_shutdown = self._get_counter(
        'sandbox_shutdown',
        description='Sandbox shutdown counter',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'error': str
        }
    )
    self._sandbox_count = self._get_scalar(
        'sandbox_count',
        description='Sandbox count',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'status': str,
        },
    )
    self._sandbox_housekeep = self._get_counter(
        'sandbox_housekeep',
        description='Sandbox housekeeping counter',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'error': str,
        }
    )
    self._sandbox_activity = self._get_counter(
        'sandbox_activity',
        description='Sandbox activity counter',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'activity': str,
            'error': str,
        }
    )

    # Sandbox scalars.
    self._sandbox_lifetime_ms = self._get_distribution(
        'sandbox_lifetime_ms',
        description='Sandbox life time in milliseconds',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'error': str,
        }
    )
    self._sandbox_start_duration_ms = self._get_distribution(
        'sandbox_start_duration_ms',
        description='Sandbox start duration in milliseconds',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'error': str,
        }
    )
    self._sandbox_shutdown_duration_ms = self._get_distribution(
        'sandbox_shutdown_duration_ms',
        description='Sandbox shutdown duration in milliseconds',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'error': str,
        }
    )
    self._sandbox_housekeep_duration_ms = self._get_distribution(
        'sandbox_housekeep_duration_ms',
        description='Sandbox housekeeping duration in milliseconds',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'error': str,
        }
    )
    self._sandbox_status_duration_ms = self._get_distribution(
        'sandbox_status_duration_ms',
        description='Sandbox duration of specific status in milliseconds',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'status': str,
        }
    )
    self._sandbox_activity_duration_ms = self._get_distribution(
        'sandbox_activity_duration_ms',
        description='Sandbox activity duration in milliseconds',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'activity': str,
            'error': str,
        }
    )

    #
    # Sandbox session metrics.
    #

    self._sandbox_session_start_duration_ms = self._get_distribution(
        'sandbox_session_start_duration_ms',
        description='Sandbox session start duration in milliseconds',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'error': str,
        }
    )
    self._sandbox_session_end_duration_ms = self._get_distribution(
        'sandbox_session_end_duration_ms',
        description='Sandbox session end duration in milliseconds',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'error': str,
        }
    )
    self._sandbox_session_lifetime_ms = self._get_distribution(
        'sandbox_session_lifetime_ms',
        description='Sandbox session lifetime in milliseconds',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'error': str,
        }
    )

    #
    # Feature metrics.
    #

    # Feature counters.
    self._feature_setup = self._get_counter(
        'feature_setup',
        description='Feature setup counter',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'feature_name': str,
            'error': str,
        }
    )
    self._feature_teardown = self._get_counter(
        'feature_teardown',
        description='Feature teardown counter',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'feature_name': str,
            'error': str,
        }
    )
    self._feature_setup_session = self._get_counter(
        'feature_setup_session',
        description='Feature setup session counter',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'feature_name': str,
            'error': str,
        }
    )
    self._feature_teardown_session = self._get_counter(
        'feature_teardown_session',
        description='Feature teardown session counter',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'feature_name': str,
            'error': str,
        }
    )
    self._feature_activity = self._get_counter(
        'feature_activity',
        description='Feature activity counter',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'activity': str,
            'error': str,
        }
    )
    self._feature_housekeep = self._get_counter(
        'feature_housekeep',
        description='Feature housekeeping counter',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'feature_name': str,
            'error': str,
        }
    )

    # Feature scalars.
    self._feature_setup_duration_ms = self._get_distribution(
        'feature_setup_duration_ms',
        description='Feature setup duration in milliseconds',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'feature_name': str,
            'error': str,
        }
    )
    self._feature_teardown_duration_ms = self._get_distribution(
        'feature_teardown_duration_ms',
        description='Feature teardown duration in milliseconds',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'feature_name': str,
            'error': str,
        }
    )
    self._feature_setup_session_duration_ms = self._get_distribution(
        'feature_setup_session_duration_ms',
        description='Feature setup session duration in milliseconds',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'feature_name': str,
            'error': str,
        }
    )
    self._feature_teardown_session_duration_ms = self._get_distribution(
        'feature_teardown_session_duration_ms',
        description='Feature teardown session duration in milliseconds',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'feature_name': str,
            'error': str,
        }
    )
    self._feature_activity_duration_ms = self._get_distribution(
        'feature_activity_duration_ms',
        description='Feature activity duration in milliseconds',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'activity': str,
            'error': str,
        }
    )
    self._feature_housekeep_duration_ms = self._get_distribution(
        'feature_housekeep_duration_ms',
        description='Feature housekeeping duration in milliseconds',
        parameters={
            'app': str,
            'environment_id': str,
            'image_id': str,
            'feature_name': str,
            'error': str,
        }
    )

  def on_environment_starting(
      self,
      environment: interface.Environment
  ) -> None:
    """Called when the environment is starting."""
    self._initialize_metrics()

  def on_environment_housekeep(
      self,
      environment: interface.Environment,
      counter: int,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when the environment is housekeeping."""
    self._environment_housekeep_duration_ms.record(
        int(duration * 1000),
        app=self.app,
        environment_id=str(environment.id),
        error=self._error_tag(error)
    )

  def on_sandbox_start(
      self,
      sandbox: interface.Sandbox,
      duration: float,
      error: BaseException | None
  ) -> None:
    self._sandbox_start.increment(
        app=self.app,
        environment_id=str(sandbox.environment.id),
        image_id=sandbox.image_id,
        error=self._error_tag(error)
    )
    self._sandbox_start_duration_ms.record(
        int(duration * 1000),
        app=self.app,
        environment_id=str(sandbox.environment.id),
        image_id=sandbox.image_id,
        error=self._error_tag(error)
    )

  def on_sandbox_status_change(
      self,
      sandbox: interface.Sandbox,
      old_status: interface.Sandbox.Status,
      new_status: interface.Sandbox.Status,
      span: float
  ) -> None:
    self._sandbox_status_duration_ms.record(
        int(span * 1000),
        app=self.app,
        environment_id=str(sandbox.environment.id),
        image_id=sandbox.image_id,
        status=old_status.value
    )
    if old_status != interface.Sandbox.Status.CREATED:
      self._sandbox_count.increment(
          delta=-1,
          app=self.app,
          environment_id=str(sandbox.environment.id),
          image_id=sandbox.image_id,
          status=old_status.value
      )
    if new_status != interface.Sandbox.Status.OFFLINE:
      self._sandbox_count.increment(
          app=self.app,
          environment_id=str(sandbox.environment.id),
          image_id=sandbox.image_id,
          status=new_status.value
      )

  def on_sandbox_shutdown(
      self,
      sandbox: interface.Sandbox,
      duration: float,
      lifetime: float,
      error: BaseException | None
  ) -> None:
    self._sandbox_shutdown.increment(
        app=self.app,
        environment_id=str(sandbox.environment.id),
        image_id=sandbox.image_id,
        error=self._error_tag(error)
    )
    self._sandbox_shutdown_duration_ms.record(
        int(duration * 1000),
        app=self.app,
        environment_id=str(sandbox.environment.id),
        image_id=sandbox.image_id,
        error=self._error_tag(error)
    )
    self._sandbox_lifetime_ms.record(
        int(lifetime * 1000),
        app=self.app,
        environment_id=str(sandbox.environment.id),
        image_id=sandbox.image_id,
        error=self._error_tag(error)
    )

  def on_sandbox_session_start(
      self,
      sandbox: interface.Sandbox,
      session_id: str,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox session starts."""
    self._sandbox_session_start_duration_ms.record(
        int(duration * 1000),
        app=self.app,
        environment_id=str(sandbox.environment.id),
        image_id=sandbox.image_id,
        error=self._error_tag(error)
    )

  def on_sandbox_session_end(
      self,
      sandbox: interface.Sandbox,
      session_id: str,
      duration: float,
      lifetime: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox session ends."""
    self._sandbox_session_end_duration_ms.record(
        int(duration * 1000),
        app=self.app,
        environment_id=str(sandbox.environment.id),
        image_id=sandbox.image_id,
        error=self._error_tag(error)
    )
    self._sandbox_session_lifetime_ms.record(
        int(lifetime * 1000),
        app=self.app,
        environment_id=str(sandbox.environment.id),
        image_id=sandbox.image_id,
        error=self._error_tag(error)
    )

  def on_sandbox_activity(
      self,
      name: str,
      sandbox: interface.Sandbox,
      session_id: str | None,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when a sandbox activity is performed."""
    self._sandbox_activity.increment(
        app=self.app,
        environment_id=str(sandbox.environment.id),
        image_id=sandbox.image_id,
        activity=name,
        error=self._error_tag(error)
    )
    self._sandbox_activity_duration_ms.record(
        int(duration * 1000),
        app=self.app,
        environment_id=str(sandbox.environment.id),
        image_id=sandbox.image_id,
        activity=name,
        error=self._error_tag(error)
    )

  def on_sandbox_housekeep(
      self,
      sandbox: interface.Sandbox,
      counter: int,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when a sandbox feature is housekeeping."""
    self._sandbox_housekeep.increment(
        app=self.app,
        environment_id=str(sandbox.environment.id),
        image_id=sandbox.image_id,
        error=self._error_tag(error)
    )
    self._sandbox_housekeep_duration_ms.record(
        int(duration * 1000),
        app=self.app,
        environment_id=str(sandbox.environment.id),
        image_id=sandbox.image_id,
        error=self._error_tag(error)
    )

  def on_feature_setup(
      self,
      feature: interface.Feature,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is setup."""
    image_id = feature.sandbox.image_id if feature.sandbox else ''
    self._feature_setup.increment(
        app=self.app,
        environment_id=str(feature.environment.id),
        image_id=image_id,
        feature_name=feature.name,
        error=self._error_tag(error)
    )
    self._feature_setup_duration_ms.record(
        int(duration * 1000),
        app=self.app,
        environment_id=str(feature.environment.id),
        image_id=image_id,
        feature_name=feature.name,
        error=self._error_tag(error)
    )

  def on_feature_teardown(
      self,
      feature: interface.Feature,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is teardown."""
    image_id = feature.sandbox.image_id if feature.sandbox else ''
    self._feature_teardown.increment(
        app=self.app,
        environment_id=str(feature.environment.id),
        image_id=image_id,
        feature_name=feature.name,
        error=self._error_tag(error)
    )
    self._feature_teardown_duration_ms.record(
        int(duration * 1000),
        app=self.app,
        environment_id=str(feature.environment.id),
        image_id=image_id,
        feature_name=feature.name,
        error=self._error_tag(error)
    )

  def on_feature_setup_session(
      self,
      feature: interface.Feature,
      session_id: str | None,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is setup."""
    image_id = feature.sandbox.image_id if feature.sandbox else ''
    self._feature_setup_session.increment(
        app=self.app,
        environment_id=str(feature.environment.id),
        image_id=image_id,
        feature_name=feature.name,
        error=self._error_tag(error)
    )
    self._feature_setup_session_duration_ms.record(
        int(duration * 1000),
        app=self.app,
        environment_id=str(feature.environment.id),
        image_id=image_id,
        feature_name=feature.name,
        error=self._error_tag(error)
    )

  def on_feature_teardown_session(
      self,
      feature: interface.Feature,
      session_id: str,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is teardown."""
    image_id = feature.sandbox.image_id if feature.sandbox else ''
    self._feature_teardown_session.increment(
        app=self.app,
        environment_id=str(feature.environment.id),
        image_id=image_id,
        feature_name=feature.name,
        error=self._error_tag(error)
    )
    self._feature_teardown_session_duration_ms.record(
        int(duration * 1000),
        app=self.app,
        environment_id=str(feature.environment.id),
        image_id=image_id,
        feature_name=feature.name,
        error=self._error_tag(error)
    )

  def on_feature_activity(
      self,
      name: str,
      feature: interface.Feature,
      session_id: str | None,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when a feature activity is performed."""
    image_id = feature.sandbox.image_id if feature.sandbox else ''
    self._feature_activity.increment(
        app=self.app,
        environment_id=str(feature.environment.id),
        image_id=image_id,
        activity=name,
        error=self._error_tag(error)
    )
    self._feature_activity_duration_ms.record(
        int(duration * 1000),
        app=self.app,
        environment_id=str(feature.environment.id),
        image_id=image_id,
        activity=name,
        error=self._error_tag(error)
    )

  def on_feature_housekeep(
      self,
      feature: interface.Feature,
      counter: int,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when a sandbox feature is housekeeping."""
    image_id = feature.sandbox.image_id if feature.sandbox else ''
    self._feature_housekeep.increment(
        app=self.app,
        environment_id=str(feature.environment.id),
        image_id=image_id,
        feature_name=feature.name,
        error=self._error_tag(error)
    )
    self._feature_housekeep_duration_ms.record(
        int(duration * 1000),
        app=self.app,
        environment_id=str(feature.environment.id),
        image_id=image_id,
        feature_name=feature.name,
        error=self._error_tag(error)
    )



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
import unittest

from langfun.core import modality
import pyglove as pg


class CustomModality(modality.Modality):
  content: str

  def to_bytes(self):
    return self.content.encode()


class ModalityTest(unittest.TestCase):

  def test_basic(self):
    v = CustomModality('a')
    self.assertEqual(v.id, 'custom_modality:0cc175b9')
    self.assertEqual(str(v), "CustomModality(\n  content = 'a'\n)")
    self.assertEqual(v.hash, '0cc175b9')

    _ = pg.Dict(metadata=pg.Dict(x=pg.Dict(metadata=pg.Dict(y=v))))
    self.assertEqual(v.id, 'custom_modality:0cc175b9')
    self.assertEqual(str(v), "CustomModality(\n  content = 'a'\n)")
    with modality.format_modality_as_ref():
      self.assertEqual(str(v), '<<[[custom_modality:0cc175b9]]>>')

  def test_capture_rendered_modalities(self):
    x = CustomModality('a')
    y = CustomModality('b')
    z = CustomModality('b')

    with modality.capture_rendered_modalities() as rendered_modalities:
      with modality.format_modality_as_ref():
        self.assertEqual(
            f'Hello {x} {y} {z}',
            (
                'Hello <<[[custom_modality:0cc175b9]]>> '
                '<<[[custom_modality:92eb5ffe]]>> '
                '<<[[custom_modality:92eb5ffe]]>>'
            )
        )
    self.assertEqual(len(rendered_modalities), 2)
    self.assertIs(rendered_modalities['custom_modality:0cc175b9'].value, x)
    # y and z share the same content will be treated as the same object.
    self.assertIs(rendered_modalities['custom_modality:92eb5ffe'].value, z)


class ModalityRefTest(unittest.TestCase):

  def test_placehold_and_restore(self):
    class A(pg.Object):
      x: Any
      y: Any

    image_a = CustomModality('a')
    image_b = CustomModality('b')
    a = A(x=dict(z=image_a), y=image_b)
    a_placehold = modality.ModalityRef.placehold(a)
    self.assertEqual(
        a_placehold,
        A(x=dict(z=modality.ModalityRef(image_a.id)),
          y=modality.ModalityRef(image_b.id)),
    )
    a_restore = modality.ModalityRef.restore(
        a_placehold.clone(),
        {image_a.id: image_a, image_b.id: image_b},
    )
    self.assertTrue(pg.eq(a_restore, a))
    self.assertEqual(
        modality.ModalityRef.placehold(a.x),
        dict(z=modality.ModalityRef(image_a.id)),
    )
    with self.assertRaisesRegex(ValueError, 'Modality .* not found'):
      modality.ModalityRef.restore(a_placehold, {image_a.id: image_a})

  def test_from_value(self):
    class A(pg.Object):
      x: Any
      y: Any

    a = A(x=dict(z=CustomModality('a')), y=CustomModality('b'))
    self.assertTrue(
        pg.eq(
            modality.Modality.from_value(a),
            {
                'custom_modality:0cc175b9': CustomModality('a'),
                'custom_modality:92eb5ffe': CustomModality('b'),
            },
        )
    )
    self.assertTrue(
        pg.eq(
            modality.Modality.from_value(a.x.z),
            {
                'custom_modality:0cc175b9': CustomModality('a'),
            },
        )
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python-based prompting protocol."""

import inspect
import io
import re
import sys
import textwrap
import typing
from typing import Any, Sequence, Type
import langfun.core as lf
from langfun.core.coding.python import correction
from langfun.core.structured.schema import base
import pyglove as pg


class PythonPromptingProtocol(base.PromptingProtocol):
  """Python-based prompting protocol."""

  NAME = 'python'

  def schema_repr(
      self,
      schema: base.Schema,
      *,
      include_result_definition: bool = True,
      markdown: bool = True,
      **kwargs,
  ) -> str:
    ret = ''
    if include_result_definition:
      ret += self.result_definition(schema)
    class_definition_str = self.class_definitions(
        schema, markdown=markdown, **kwargs
    )
    if class_definition_str:
      ret += f'\n\n{class_definition_str}'
    return ret.strip()

  def class_definitions(
      self,
      schema: base.Schema,
      additional_dependencies: list[Type[Any]] | None = None,
      **kwargs
  ) -> str | None:
    """Returns a string containing of class definitions from a schema."""
    deps = schema.class_dependencies(
        include_base_classes=False, include_subclasses=True
    )
    allowed_dependencies = set(deps)
    if additional_dependencies:
      allowed_dependencies.update(additional_dependencies)
    return class_definitions(
        deps, allowed_dependencies=allowed_dependencies, **kwargs
    )

  def result_definition(self, schema: base.Schema) -> str:
    return base.annotation(schema.spec)

  def value_repr(
      self,
      value: Any,
      schema: base.Schema | None = None,
      *,
      compact: bool = True,
      verbose: bool = False,
      markdown: bool = True,
      assign_to_var: str | None = None,
      **kwargs) -> str:
    del schema
    if inspect.isclass(value):
      cls_schema = base.Schema.from_value(value)
      if isinstance(cls_schema.spec, pg.typing.Object):
        object_code = self.class_definitions(
            cls_schema,
            markdown=markdown,
            # We add `pg.Object` as additional dependencies to the class
            # definition so exemplars for class generation could show
            # pg.Object as their bases.
            additional_dependencies=[pg.Object]
        )
        assert object_code is not None
        return object_code
      else:
        object_code = self.result_definition(cls_schema)
    elif isinstance(value, lf.Template):
      return str(value)
    else:
      object_code = pg.format(
          value, compact=compact, verbose=verbose, python_format=True
      )
      if assign_to_var is not None:
        object_code = f'{assign_to_var} = {object_code}'
    if markdown:
      return f'```python\n{object_code}\n```'
    return object_code

  def parse_value(
      self,
      text: str,
      schema: base.Schema | None = None,
      *,
      additional_context: dict[str, Type[Any]] | None = None,
      permission: pg.coding.CodePermission = (
          pg.coding.CodePermission.ASSIGN | pg.coding.CodePermission.CALL
      ),
      autofix=0,
      autofix_lm: lf.LanguageModel = lf.contextual(),
      **kwargs,
  ) -> Any:
    """Parses a Python string into a structured object."""
    del kwargs
    global_vars = additional_context or {}
    if schema is not None:
      dependencies = schema.class_dependencies()
      global_vars.update({d.__name__: d for d in dependencies})
    return structure_from_python(
        text,
        global_vars=global_vars,
        autofix=autofix,
        autofix_lm=autofix_lm,
        permission=permission,
    )


def structure_from_python(
    code: str,
    *,
    global_vars: dict[str, Any] | None = None,
    permission: pg.coding.CodePermission = (
        pg.coding.CodePermission.ASSIGN | pg.coding.CodePermission.CALL
    ),
    autofix=0,
    autofix_lm: lf.LanguageModel = lf.contextual(),
) -> Any:
  """Evaluates structure from Python code with access to symbols."""
  global_vars = global_vars or {}
  global_vars.update({
      'pg': pg,
      'Object': pg.Object,
      'Any': typing.Any,
      'List': typing.List,
      'Tuple': typing.Tuple,
      'Dict': typing.Dict,
      'Sequence': typing.Sequence,
      'Optional': typing.Optional,
      'Union': typing.Union,
      # Special value markers.
      'UNKNOWN': base.UNKNOWN,
  })
  # We are creating objects here, so we execute the code without a sandbox.
  return correction.run_with_correction(
      code,
      global_vars=global_vars,
      sandbox=False,
      max_attempts=autofix,
      lm=autofix_lm,
      permission=permission,
  )


def source_form(value, compact: bool = True, markdown: bool = False) -> str:
  """Returns the source code form of an object."""
  return PythonPromptingProtocol().value_repr(
      value, compact=compact, markdown=markdown
  )


def include_method_in_prompt(method):
  """Decorator to include a method in the class definition of the prompt."""
  setattr(method, '__show_in_prompt__', True)
  return method


def should_include_method_in_prompt(method):
  """Returns True if the method should be shown in the prompt."""
  return getattr(method, '__show_in_prompt__', False)


def class_definition(
    cls,
    strict: bool = False,
    allowed_dependencies: set[Type[Any]] | None = None,
) -> str:
  """Returns the Python class definition."""
  out = io.StringIO()
  schema = pg.schema(cls)
  eligible_bases = []
  for base_cls in cls.__bases__:
    if base_cls is not object:
      if allowed_dependencies is None or base_cls in allowed_dependencies:
        eligible_bases.append(base_cls.__name__)

  if eligible_bases:
    base_cls_str = ', '.join(eligible_bases)
    out.write(f'class {cls.__name__}({base_cls_str}):\n')
  else:
    out.write(f'class {cls.__name__}:\n')

  if cls.__doc__:
    doc_lines = cls.__doc__.strip().split('\n')
    if len(doc_lines) == 1:
      out.write(f'  """{cls.__doc__}"""\n')
    else:
      out.write('  """')

      # Since Python 3.13, the indentation of docstring lines is removed.
      # Therefore, we add two spaces to each non-empty line to keep the
      # indentation consistent with the class definition.
      if sys.version_info >= (3, 13):
        for i in range(1, len(doc_lines)):
          if doc_lines[i]:
            doc_lines[i] = ' ' * 2 + doc_lines[i]

      for line in doc_lines:
        out.write(line)
        out.write('\n')
      out.write('  """\n')

  empty_class = True
  if schema.fields:
    for key, field in schema.items():
      if not isinstance(key, pg.typing.ConstStrKey):
        pg.logging.warning(
            'Variable-length keyword arguments is not supported in '
            f'structured parsing or query. Encountered: {cls}, Schema: {schema}'
        )
        continue

      # Skip fields that are marked as excluded from the prompt sent to LLM
      # for OOP.
      if field.metadata.get('exclude_from_prompt', False):
        continue

      # Write field doc string as comments before the field definition.
      if field.description:
        for line in field.description.split('\n'):
          if line:
            out.write('  # ')
            out.write(line)
            out.write('\n')

      annotation_str = base.annotation(
          field.value, strict=strict, allowed_dependencies=allowed_dependencies
      )
      out.write(f'  {field.key}: {annotation_str}')
      out.write('\n')
      empty_class = False

  for method in _iter_newly_defined_methods(cls, allowed_dependencies):
    source = inspect.getsource(method)
    # Remove decorators from the method definition.
    source = re.sub(r'\s*@.*\.include_method_in_prompt.*\n', '', source)
    out.write('\n')
    out.write(
        textwrap.indent(
            inspect.cleandoc('\n' + source), ' ' * 2)
    )
    out.write('\n')
    empty_class = False

  if empty_class:
    out.write('  pass\n')
  return out.getvalue()


def _iter_newly_defined_methods(
    cls, allowed_dependencies: set[Type[Any]] | None):
  names = {attr_name: True for attr_name in dir(cls)}
  for base_cls in cls.__bases__:
    if allowed_dependencies is None or base_cls in allowed_dependencies:
      for name in dir(base_cls):
        names.pop(name, None)
  for name in names.keys():
    attr = getattr(cls, name)
    if callable(attr) and should_include_method_in_prompt(attr):
      yield attr


def class_definitions(
    classes: Sequence[Type[Any]],
    *,
    allowed_dependencies: set[Type[Any]] | None = None,
    strict: bool = False,
    markdown: bool = False,
) -> str | None:
  """Returns a string for class definitions."""
  if not classes:
    return None
  def_str = io.StringIO()
  for i, cls in enumerate(classes):
    if i > 0:
      def_str.write('\n')
    def_str.write(
        class_definition(
            cls,
            strict=strict,
            allowed_dependencies=allowed_dependencies,
        )
    )
  ret = def_str.getvalue()
  if markdown and ret:
    ret = f'```python\n{ret}```'
  return ret



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for debug runner."""
import os
import tempfile
from typing import Any
import unittest

from langfun.core.eval.v2 import eval_test_helper
from langfun.core.eval.v2.runners import debug  # pylint: disable=unused-import

import pyglove as pg


class DebugRunnerTest(unittest.TestCase):

  def assert_same_list(self, actual: list[Any], expected: list[Any]):
    self.assertEqual(len(actual), len(expected))
    for i, (x, y) in enumerate(zip(actual, expected)):
      if x is not y:
        print(i, pg.diff(x, y))
      self.assertIs(x, y)

  def test_debug_runner(self):
    plugin = eval_test_helper.TestPlugin()
    exp = eval_test_helper.test_experiment()
    root_dir = os.path.join(tempfile.mkdtemp(), 'test_debug_runner')
    run = exp.run(root_dir, runner='debug', plugins=[plugin])

    self.assertIsNotNone(plugin.start_time)
    self.assertIsNotNone(plugin.complete_time)
    self.assertGreater(plugin.complete_time, plugin.start_time)

    self.assertEqual(
        len(plugin.started_experiments), len(exp.nodes)
    )
    self.assertEqual(
        len(plugin.completed_experiments), len(exp.nodes)
    )
    self.assertEqual(
        len(plugin.started_example_ids), 6 * 1
    )
    self.assertEqual(
        len(plugin.completed_example_ids), 6 * 1
    )
    self.assert_same_list(plugin.skipped_experiments, [])
    self.assertFalse(
        pg.io.path_exists(os.path.join(run.output_root, 'run.json'))
    )

    for node in exp.nodes:
      self.assertTrue(node.progress.is_started)
      self.assertTrue(node.progress.is_completed)
      if node.is_leaf:
        self.assertEqual(node.progress.num_skipped, 0)
        self.assertEqual(node.progress.num_completed, 1)
        self.assertEqual(node.progress.num_failed, 0)
      else:
        self.assertEqual(node.progress.num_skipped, 0)
        self.assertEqual(node.progress.num_failed, 0)
        self.assertEqual(node.progress.num_processed, node.progress.num_total)


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from langfun.env import interface
from langfun.env.event_handlers import chain


class MockEventHandler(interface.EventHandler):

  def __init__(self):
    self.calls = []

  def _record_call(self, method_name, *args, **kwargs):
    self.calls.append((method_name, args, kwargs))

  def on_environment_starting(self, environment):
    self._record_call('on_environment_starting', environment)

  def on_environment_shutting_down(self, environment, offline_duration):
    self._record_call(
        'on_environment_shutting_down', environment, offline_duration
    )

  def on_environment_start(self, environment, duration, error):
    self._record_call('on_environment_start', environment, duration, error)

  def on_environment_housekeep(
      self, environment, counter, duration, error, **kwargs
  ):
    self._record_call(
        'on_environment_housekeep',
        environment, counter, duration, error, **kwargs
    )

  def on_environment_shutdown(self, environment, duration, lifetime, error):
    self._record_call(
        'on_environment_shutdown', environment, duration, lifetime, error
    )

  def on_sandbox_start(self, sandbox, duration, error):
    self._record_call('on_sandbox_start', sandbox, duration, error)

  def on_sandbox_status_change(
      self, sandbox, old_status, new_status, span
  ):
    self._record_call(
        'on_sandbox_status_change',
        sandbox,
        old_status,
        new_status,
        span,
    )

  def on_sandbox_shutdown(
      self, sandbox, duration, lifetime, error
  ):
    self._record_call('on_sandbox_shutdown', sandbox, duration, lifetime, error)

  def on_sandbox_session_start(self, sandbox, session_id, duration, error):
    self._record_call(
        'on_sandbox_session_start', sandbox, session_id, duration, error
    )

  def on_sandbox_session_end(
      self, sandbox, session_id, duration, lifetime, error
  ):
    self._record_call(
        'on_sandbox_session_end',
        sandbox,
        session_id,
        duration,
        lifetime,
        error,
    )

  def on_sandbox_activity(
      self,
      name,
      sandbox,
      session_id,
      duration,
      error,
      **kwargs,
  ):
    self._record_call(
        'on_sandbox_activity',
        name,
        sandbox,
        session_id,
        duration,
        error,
        **kwargs,
    )

  def on_sandbox_housekeep(self, sandbox, counter, duration, error, **kwargs):
    self._record_call(
        'on_sandbox_housekeep',
        sandbox,
        counter,
        duration,
        error,
        **kwargs,
    )

  def on_feature_setup(self, feature, duration, error):
    self._record_call('on_feature_setup', feature, duration, error)

  def on_feature_teardown(self, feature, duration, error):
    self._record_call('on_feature_teardown', feature, duration, error)

  def on_feature_setup_session(self, feature, session_id, duration, error):
    self._record_call(
        'on_feature_setup_session',
        feature,
        session_id,
        duration,
        error,
    )

  def on_feature_teardown_session(self, feature, session_id, duration, error):
    self._record_call(
        'on_feature_teardown_session',
        feature,
        session_id,
        duration,
        error,
    )

  def on_feature_activity(
      self, name, feature, session_id, duration, error, **kwargs,
  ):
    self._record_call(
        'on_feature_activity',
        name,
        feature,
        session_id,
        duration,
        error,
        **kwargs,
    )

  def on_feature_housekeep(
      self, feature, counter, duration, error, **kwargs
  ):
    self._record_call(
        'on_feature_housekeep',
        feature,
        counter,
        duration,
        error,
        **kwargs,
    )


class EventHandlerChainTest(unittest.TestCase):

  def test_chain(self):
    handler1 = MockEventHandler()
    handler2 = MockEventHandler()
    chain_handler = chain.EventHandlerChain([handler1, handler2])

    env = object()
    sandbox = object()
    feature = object()

    chain_handler.on_environment_starting(env)
    chain_handler.on_environment_shutting_down(env, 1.0)
    chain_handler.on_environment_start(env, 2.0, None)
    chain_handler.on_environment_housekeep(env, 1, 3.0, None, a=1)
    chain_handler.on_environment_shutdown(env, 4.0, 5.0, None)
    chain_handler.on_sandbox_start(sandbox, 6.0, None)
    chain_handler.on_sandbox_status_change(sandbox, 'old', 'new', 7.0)
    chain_handler.on_sandbox_shutdown(sandbox, 8.0, 9.0, None)
    chain_handler.on_sandbox_session_start(sandbox, 's2', 16.0, None)
    chain_handler.on_sandbox_session_end(sandbox, 's2', 17.0, 18.0, None)
    chain_handler.on_sandbox_activity('act', sandbox, 's2', 19.0, None, d=4)
    chain_handler.on_sandbox_housekeep(sandbox, 2, 10.0, None, b=2)
    chain_handler.on_feature_setup(feature, 11.0, None)
    chain_handler.on_feature_teardown(feature, 12.0, None)
    chain_handler.on_feature_setup_session(feature, 's1', 13.0, None)
    chain_handler.on_feature_teardown_session(feature, 's1', 14.0, None)
    chain_handler.on_feature_activity('act', feature, 's1', 16.0, None, d=5)
    chain_handler.on_feature_housekeep(feature, 3, 15.0, None, c=3)

    self.assertEqual(handler1.calls, handler2.calls)
    self.assertEqual(
        handler1.calls,
        [
            ('on_environment_starting', (env,), {}),
            ('on_environment_shutting_down', (env, 1.0), {}),
            ('on_environment_start', (env, 2.0, None), {}),
            ('on_environment_housekeep', (env, 1, 3.0, None), {'a': 1}),
            ('on_environment_shutdown', (env, 4.0, 5.0, None), {}),
            ('on_sandbox_start', (sandbox, 6.0, None), {}),
            ('on_sandbox_status_change', (sandbox, 'old', 'new', 7.0), {}),
            ('on_sandbox_shutdown', (sandbox, 8.0, 9.0, None), {}),
            ('on_sandbox_session_start', (sandbox, 's2', 16.0, None), {}),
            ('on_sandbox_session_end', (sandbox, 's2', 17.0, 18.0, None), {}),
            (
                'on_sandbox_activity',
                ('act', sandbox, 's2', 19.0, None),
                {'d': 4},
            ),
            ('on_sandbox_housekeep', (sandbox, 2, 10.0, None), {'b': 2}),
            ('on_feature_setup', (feature, 11.0, None), {}),
            ('on_feature_teardown', (feature, 12.0, None), {}),
            ('on_feature_setup_session', (feature, 's1', 13.0, None), {}),
            ('on_feature_teardown_session', (feature, 's1', 14.0, None), {}),
            (
                'on_feature_activity',
                ('act', feature, 's1', 16.0, None),
                {'d': 5}
            ),
            ('on_feature_housekeep', (feature, 3, 15.0, None), {'c': 3}),
        ],
    )

  def test_add_remove(self):
    handler1 = MockEventHandler()
    handler2 = MockEventHandler()
    chain_handler = chain.EventHandlerChain([handler1])
    chain_handler.add(handler2)
    env = object()
    chain_handler.on_environment_starting(env)
    self.assertEqual(handler1.calls, [('on_environment_starting', (env,), {})])
    self.assertEqual(handler2.calls, [('on_environment_starting', (env,), {})])

    chain_handler.remove(handler1)
    chain_handler.on_environment_start(env, 1.0, None)
    self.assertEqual(handler1.calls, [('on_environment_starting', (env,), {})])
    self.assertEqual(
        handler2.calls,
        [
            ('on_environment_starting', (env,), {}),
            ('on_environment_start', (env, 1.0, None), {}),
        ],
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common base class for sandboxes.

This module provides an base class `BaseSandbox` for sandboxes, which takes care
of the sandbox lifecycle and housekeeping. It also provides a decorator for
sandbox service methods to handle errors and trigger shutdowns. Please note that
this base class is intended to provide a convenient way to implement sandboxes,
and not all sandbox implementations need to subclass it. Also `BaseSandbox`
is not coupled with `BaseEnvironment` and `BaseFeature`, and is expected to
work with the `Environment` and `Feature` interfaces directly.
"""

import contextlib
import functools
import threading
import time
from typing import Annotated, Any, Iterator

from langfun.env import interface
import pyglove as pg


class BaseSandbox(interface.Sandbox):
  """Base class for a sandbox."""

  id: Annotated[
      interface.Sandbox.Id,
      'The identifier for the sandbox.'
  ]

  image_id: Annotated[
      str,
      'The image id for the sandbox.'
  ]

  environment: Annotated[
      pg.Ref[interface.Environment],
      'The parent environment.'
  ]

  reusable: Annotated[
      bool,
      (
          'If True, the sandbox can be reused for multiple user sessions at '
          'different times.'
      )
  ] = False

  keepalive_interval: Annotated[
      float | None,
      'Interval to ping the sandbox for keeping it alive..'
  ] = 60.0

  proactive_session_setup: Annotated[
      bool,
      (
          'If True, the sandbox will perform setup work before a user session '
          'is started. This is useful for sandboxes that need to perform '
          'heavy setup work, which could block the user thread for a long '
          'time. Applicable only when `reusable` is True.'
      )
  ] = True

  #
  # There is no required methods that subclasses must implement.
  # Subclasses can override the following methods:
  #

  def _start(self) -> None:
    """Implementation of start(). Subclasses can override.

    Raises:
      interface.SandboxStateError: If the sandbox is in a bad state.
    """

  def _shutdown(self) -> None:
    """Implementation of shutdown(). Subclasses can override.

    Raises:
      interface.SandboxStateError: If the sandbox is in a bad state.
    """

  def _set_status(self, status: interface.Sandbox.Status) -> None:
    """Sets the status of the sandbox."""
    assert self._status != status, (self._status, status)
    self.on_status_change(self._status, status)
    self._status = status
    self._status_start_time = time.time()

  def report_state_error(self, e: interface.SandboxStateError) -> None:
    """Reports sandbox state errors."""
    if e not in self._state_errors:
      self._state_errors.append(e)

  def _setup_features(self) -> None:
    """Starts the features in the sandbox."""
    # We keep track of the features that have setup called so we can teardown
    # the features when the sandbox is shutdown.
    self._features_with_setup_called.clear()

    for feature in self._features.values():
      self._features_with_setup_called.add(feature.name)
      feature.setup(self)

  def _setup_session(self) -> None:
    """Sets up a new session for the sandbox."""
    # We keep track of the features that have setup_session called so we can
    # teardown the session for them when the session ends.
    self._features_with_setup_session_called.clear()

    for feature in self._features.values():
      self._features_with_setup_session_called.add(feature.name)
      feature.setup_session()

  def _teardown_features(self) -> interface.FeatureTeardownError | None:
    """Tears down the features in the sandbox.

    IMPORTANT: This method shall not raise any exceptions.

    Returns:
      FeatureTeardownError: If feature teardown failed with errors.
        Otherwise None.
    """
    errors = {}
    for feature in self._features.values():
      if feature.name in self._features_with_setup_called:
        try:
          feature.teardown()
        except BaseException as e:  # pylint: disable=broad-except
          if isinstance(e, interface.SandboxStateError):
            self.report_state_error(e)
          errors[feature.name] = e
    if errors:
      return interface.FeatureTeardownError(sandbox=self, errors=errors)
    return None

  def _start_session(self) -> None:
    """Starts a user session.

    Raises:
      BaseException: If feature setup failed with user-defined errors.
      SandboxStateError: If sandbox or feature setup fail due sandbox state
        errors.
    """
    # When pre-session setup is enabled, the session setup is performed
    # before the session is started. Otherwise we setup the session here.
    if not self._enable_pre_session_setup:
      self._setup_session()

  def _end_session(self) -> interface.SessionTeardownError | None:
    """Ends a user session.

    IMPORTANT: This method shall not raise any exceptions.

    Returns:
      SessionTeardownError: If session teardown failed with errors.
        Otherwise None.
    """
    feature_teardown_errors = {}
    for name, feature in self._features.items():
      if name in self._features_with_setup_session_called:
        try:
          feature.teardown_session()
        except BaseException as e:  # pylint: disable=broad-except
          if isinstance(e, interface.SandboxStateError):
            self.report_state_error(e)
          feature_teardown_errors[name] = e

    return interface.SessionTeardownError(
        sandbox=self, errors=feature_teardown_errors
    ) if feature_teardown_errors else None

  def _ping(self) -> None:
    """Implementation of ping for health checking."""

  #
  # Sandbox basics.
  #

  def _on_bound(self) -> None:
    """Called when the sandbox is bound."""
    super()._on_bound()
    self._features = pg.Dict({
        name: pg.clone(feature)
        for name, feature in self.environment.features.items()
        if feature.is_applicable(self.image_id)
    })
    self._event_handler = self.environment.event_handler
    self._enable_pre_session_setup = (
        self.reusable and self.proactive_session_setup
    )
    self._enables_housekeep = (
        self.keepalive_interval is not None
        or any(
            feature.housekeep_interval is not None
            for feature in self._features.values()
        )
    )
    self._housekeep_thread = None
    self._housekeep_counter = 0

    # Runtime state.
    self._status = self.Status.CREATED
    self._status_start_time = time.time()

    self._start_time = None
    self._state_errors = []

    self._features_with_setup_called = set()
    self._features_with_setup_session_called = set()

    self._session_id = None
    self._session_start_time = None

    # Thread local state for this sandbox.
    self._tls_state = threading.local()

  @functools.cached_property
  def working_dir(self) -> str | None:
    """Returns the working directory for the sandbox."""
    return self.id.working_dir(self.environment.root_dir)

  @property
  def status(self) -> interface.Sandbox.Status:
    """Returns the state of the sandbox."""
    return self._status

  def set_acquired(self) -> None:
    """Marks the sandbox as acquired."""
    self._set_status(self.Status.ACQUIRED)

  @property
  def housekeep_counter(self) -> int:
    """Returns the housekeeping counter."""
    return self._housekeep_counter

  @property
  def state_errors(self) -> list[interface.SandboxStateError]:
    """Returns all errors encountered during sandbox lifecycle."""
    return self._state_errors

  @property
  def is_shutting_down(self) -> bool:
    """Returns True if the sandbox is shutting down."""
    return self._status == self.Status.SHUTTING_DOWN or (
        self._state_errors and self._status == self.Status.EXITING_SESSION
    )

  @property
  def features(self) -> dict[str, interface.Feature]:
    """Returns the features in the sandbox."""
    return self._features

  #
  # Sandbox start/shutdown.
  #

  def start(self) -> None:
    """Starts the sandbox.

    State transitions:
      CREATED -> SETTING_UP -> READY: When all sandbox and feature setup
        succeeds.
      CREATED -> SETTING_UP -> SHUTTING_DOWN -> OFFLINE: When sandbox or feature
        setup fails.

    `start` and `shutdown` should be called in pairs, even when the sandbox
    fails to start. This ensures proper cleanup.

    Start may fail with two sources of errors:

    1. SandboxStateError: If sandbox or feature setup fail due to enviroment
       outage or sandbox state errors.
    2. BaseException: If feature setup failed with user-defined errors, this
       could happen when there is bug in the user code or non-environment code
       failure.

    In both cases, the sandbox will be shutdown automatically, and the error
    will be added to `errors`. The sandbox is considered dead and will not be
    further used.

    Raises:
      SandboxStateError: If the sandbox is in a bad state.
      BaseException: If feature setup failed with user-defined errors.
    """
    assert self._status == self.Status.CREATED, (
        f'Sandbox {self.id} cannot be started because '
        f'it is in {self._status} status.'
    )

    starting_time = time.time()
    self._state = self.Status.SETTING_UP

    try:
      # Start the sandbox.
      self._start()

      # Setup the features.
      self._setup_features()

      # Setup the first session if pre-session setup is enabled.
      if self._enable_pre_session_setup:
        self._setup_session()

      if self._enables_housekeep:
        self._housekeep_thread = threading.Thread(
            target=self._housekeep_loop, daemon=True
        )
        self._housekeep_thread.start()

      self._start_time = time.time()

      # Mark the sandbox as ready when all setup succeeds.
      self._set_status(self.Status.READY)

      duration = time.time() - starting_time
      self.on_start(duration)
    except BaseException as e:  # pylint: disable=broad-except
      duration = time.time() - starting_time
      pg.logging.error(
          '[%s]: Sandbox failed to start in %.2f seconds: %s',
          self.id, duration, e
      )
      if isinstance(e, interface.SandboxStateError):
        self.report_state_error(e)
      self.on_start(duration, e)
      self.shutdown()
      raise e

  def shutdown(self) -> None:
    """Shuts down the sandbox.

    State transitions:
      SHUTTING_DOWN -> SHUTTING_DOWN: No operation.
      OFFLINE -> OFFLINE: No operation.
      SETTING_UP -> SHUTTING_DOWN -> OFFLINE: When sandbox and feature
        setup fails.
      IN_SESSION -> SHUTTING_DOWN -> OFFLINE: When user session exits while
        sandbox is set not to reuse, or session teardown fails.
      FREE -> SHUTTING_DOWN -> OFFLINE: When sandbox is shutdown when the
        environment is shutting down, or housekeeping loop shuts down the
        sandbox due to housekeeping failures.

    Please be aware that `shutdown` will be called whenever an operation on the
    sandbox encounters a critical error. This means, `shutdown` should not make
    the assumption that the sandbox is in a healthy state, even `start` could
    fail. As a result, `shutdown` must allow re-entry and be thread-safe with
    other sandbox operations.

    Shutdown may fail with two sources of errors:

    1. SandboxStateError: If the sandbox is in a bad state, and feature teardown
       logic depending on a healthy sandbox may fail. In such case, we do not
       raise error to the user as the user session is considered completed. The
       sandbox is abandoned and new user sessions will be served on other
       sandboxes.

    2. BaseException: The sandbox is in good state, but user code raises error
       due to bug or non-environment code failure. In such case, errors will be
       raised to the user so the error could be surfaced and handled properly.
       The sandbox is treated as shutdown and will not be further used.

    Raises:
      BaseException: If feature teardown failed with user-defined errors.
    """

    # Allow re-entry.
    if self._status in (
        interface.Sandbox.Status.SHUTTING_DOWN,
        interface.Sandbox.Status.OFFLINE
    ):
      return

    # End current session and shutdown the sandbox if the sandbox is in session.
    if self._status == self.Status.IN_SESSION:
      self.end_session(shutdown_sandbox=True)
      return

    shutting_down_time = time.time()
    self._set_status(interface.Sandbox.Status.SHUTTING_DOWN)

    if (self._housekeep_thread is not None
        and threading.current_thread() is not self._housekeep_thread):
      self._housekeep_thread.join()
      self._housekeep_thread = None

    teardown_error = self._teardown_features()
    try:
      self._shutdown()
      self._set_status(interface.Sandbox.Status.OFFLINE)
      self.on_shutdown(
          duration=time.time() - shutting_down_time,
          error=teardown_error
      )
      shutdown_error = None
    except BaseException as e:  # pylint: disable=broad-except
      shutdown_error = e
      if isinstance(e, interface.SandboxStateError):
        self.report_state_error(e)
      self._set_status(interface.Sandbox.Status.OFFLINE)
      pg.logging.error(
          '[%s]: Sandbox shutdown with error: %s',
          self.id, e
      )
      self.on_shutdown(
          duration=time.time() - shutting_down_time,
          error=teardown_error or shutdown_error
      )

    # We raise non-state errors to the user following timely order, so the user
    # code could be surfaced and handled properly.
    if (teardown_error is not None
        and teardown_error.has_non_sandbox_state_error):
      raise teardown_error

    if shutdown_error is not None and not isinstance(
        shutdown_error, interface.SandboxStateError
    ):
      raise shutdown_error

  def ping(self) -> None:
    """Pings the sandbox to check if it is alive."""
    self._ping()

  #
  # API related to a user session.
  # A sandbox could be reused across different user sessions.
  # A user session is a sequence of stateful interactions with the sandbox,
  # Across different sessions the sandbox are considered stateless.
  #

  @property
  def session_id(self) -> str | None:
    """Returns the current user session identifier.

    session_id is set to None when the sandbox is free, and set to 'pending'
    when the sandbox is acquired by a thread, before a user session is started.

    Returns:
      The current user session identifier or None if the sandbox is not busy.
    """
    return self._session_id

  def start_session(
      self,
      session_id: str,
  ) -> None:
    """Begins a user session with the sandbox.

    State transitions:
      ACQUIRED -> SETTING_UP -> IN_SESSION: When session setup succeeds.
      ACQUIRED -> SETTING_UP -> SHUTTING_DOWN -> OFFLINE: When session setup
        fails.

    A session is a sequence of stateful interactions with the sandbox.
    Across different sessions the sandbox are considered stateless.
    `start_session` and `end_session` should always be called in pairs, even
    when the session fails to start. `Sandbox.new_session` context manager is
    the recommended way to use `start_session` and `end_session` in pairs.

    Starting a session may fail with two sources of errors:

    1. SandboxStateError: If the sandbox is in a bad state or session setup
       failed.

    2. BaseException: If session setup failed with user-defined errors.

    In both cases, the sandbox will be shutdown automatically and the
    session will be considered ended. The error will be added to `errors`.
    Future session will be served on other sandboxes.

    Args:
      session_id: The identifier for the user session.

    Raises:
      SandboxStateError: If the sandbox is already in a bad state or session
        setup failed.
      BaseException: If session setup failed with user-defined errors.
    """
    assert self._status == self.Status.ACQUIRED, (
        f'Sandbox {self.id} is not in acquired state (status={self._status}).'
    )
    assert self._session_id is None, (
        f'A user session {self._session_id} is already active '
        f'for sandbox {self.id}.'
    )
    self._set_status(self.Status.SETTING_UP)

    self._session_id = session_id
    self._session_start_time = time.time()

    try:
      self._start_session()
      self._set_status(self.Status.IN_SESSION)
      self.on_session_start(session_id, time.time() - self._session_start_time)
    except BaseException as e:  # pylint: disable=broad-except
      if isinstance(e, interface.SandboxStateError):
        self.report_state_error(e)
      self.on_session_start(
          session_id, time.time() - self._session_start_time, e
      )
      self.shutdown()
      raise e

  def end_session(self, shutdown_sandbox: bool = False) -> None:
    """Ends the user session with the sandbox.

    State transitions:
      IN_SESSION -> EXITING_SESSION -> READY: When user session exits normally,
        and sandbox is set to reuse.
      IN_SESSION -> EXITING_SESSION -> SHUTTING_DOWN -> OFFLINE: When user
        session exits while
        sandbox is set not to reuse, or session teardown fails.
      IN_SESSION -> EXITING_SESSION -> SETTING_UP -> READY: When user session
        exits normally, and sandbox is set to reuse, and proactive session setup
        is enabled.
      IN_SESSION -> EXITING_SESSION -> SETTING_UP -> SHUTTING_DOWN -> OFFLINE:
        When user session exits normally, and proactive session setup is enabled
        but fails.
      EXITING_SESSION -> EXITING_SESSION: No operation.
      not IN_SESSION -> same state: No operation

    `end_session` should always be called for each `start_session` call, even
    when the session fails to start, to ensure proper cleanup.

    `end_session` may fail with two sources of errors:

    1. SandboxStateError: If the sandbox is in a bad state or session teardown
       failed.

    2. BaseException: If session teardown failed with user-defined errors.

    In both cases, the sandbox will be shutdown automatically and the
    session will be considered ended. The error will be added to `errors`.
    Future session will be served on other sandboxes.

    However, SandboxStateError encountered during `end_session` will NOT be
    raised to the user as the user session is considered completed.

    Args:
      shutdown_sandbox: If True, the sandbox will be shutdown after session
        teardown.

    Raises:
      BaseException: If session teardown failed with user-defined errors.
    """
    if self._status == self.Status.EXITING_SESSION:
      return

    if self._status not in (
        self.Status.IN_SESSION,
    ):
      return

    assert self._session_id is not None, (
        'No user session is active for this sandbox'
    )
    # Set sandbox status to EXITING_SESSION to avoid re-entry.
    self._set_status(self.Status.EXITING_SESSION)
    shutdown_sandbox = shutdown_sandbox or not self.reusable
    ending_time = time.time()

    # Teardown features for the current session.
    end_session_error = self._end_session()
    previous_session_id = self._session_id
    self._session_id = None
    self._features_with_setup_session_called.clear()

    # If there is no state error, and proactive session setup is enabled,
    # set up the next session proactively.
    if not self.state_errors:
      if not shutdown_sandbox and self._enable_pre_session_setup:
        def _setup_next_session():
          try:
            self._setup_session()
            self._set_status(interface.Sandbox.Status.READY)
          except BaseException as e:  # pylint: disable=broad-except
            pg.logging.error(
                '[%s]: Shutting down sandbox due to practively setting up '
                'next session failed: %s',
                self.id,
                e
            )
            if isinstance(e, interface.SandboxStateError):
              self.report_state_error(e)
            self.shutdown()

        # End session before setting up the next session.
        self.on_session_end(
            previous_session_id, duration=time.time() - ending_time
        )

        # Mark the sandbox as setting up to prevent it from being acquired by
        # other threads.
        self._set_status(interface.Sandbox.Status.SETTING_UP)

        # TODO(daiyip): Consider using a thread pool to perform next session
        # setup.
        threading.Thread(target=_setup_next_session).start()
      else:
        # End session before reporting sandbox status change.
        self.on_session_end(
            previous_session_id, duration=time.time() - ending_time
        )

        # If shutdown is requested, mark the sandbox as acquired to prevent it
        # from being acquired by other threads.
        self._set_status(
            interface.Sandbox.Status.ACQUIRED if shutdown_sandbox else
            interface.Sandbox.Status.READY
        )

    # Otherwise, shutdown the sandbox.
    else:
      self.on_session_end(
          previous_session_id,
          duration=time.time() - ending_time,
          error=self.state_errors[0]
      )
      self._set_status(interface.Sandbox.Status.ACQUIRED)
      shutdown_sandbox = True

    self._session_start_time = None

    if shutdown_sandbox:
      self.shutdown()

    # We only raise errors if teardown error contains non-sandbox-state error,
    # meaning that the user code may have bug or other non-environment
    # failures.
    if (end_session_error is not None
        and end_session_error.has_non_sandbox_state_error):
      raise end_session_error  # pylint: disable=raising-bad-type

  @contextlib.contextmanager
  def track_activity(
      self,
      name: str,
      **kwargs: Any
  ) -> Iterator[None]:
    """Tracks an activity for the sandbox."""
    start_time = time.time()
    error = None
    try:
      yield None
    except BaseException as e:  # pylint: disable=broad-except
      error = e
      raise
    finally:
      self.on_activity(
          name=name,
          duration=time.time() - start_time,
          error=error,
          **kwargs
      )

  #
  # Housekeeping.
  #

  def _housekeep_loop(self) -> None:
    """Sandbox housekeeping loop."""
    now = time.time()
    last_ping = now
    last_housekeep_time = {name: now for name in self._features.keys()}

    def _next_housekeep_wait_time() -> float:
      # Decide how long to sleep for the next housekeeping.
      next_housekeep_time = None
      if self.keepalive_interval is not None:
        next_housekeep_time = last_ping + self.keepalive_interval

      for name, feature in self._features.items():
        if feature.housekeep_interval is None:
          continue
        next_feature_housekeep_time = (
            last_housekeep_time[name] + feature.housekeep_interval
        )
        if (next_housekeep_time is None
            or next_housekeep_time > next_feature_housekeep_time):
          next_housekeep_time = next_feature_housekeep_time

      # Housekeep loop is installed when at least one feature requires
      # housekeeping or the sandbox has a keepalive interval.
      assert next_housekeep_time is not None
      return max(0, next_housekeep_time - time.time())

    while self._status not in (self.Status.SHUTTING_DOWN, self.Status.OFFLINE):
      housekeep_start = time.time()
      if self.keepalive_interval is not None:
        if time.time() - last_ping > self.keepalive_interval:
          try:
            self.ping()
          except interface.SandboxStateError as e:
            pg.logging.error(
                '[%s]: Shutting down sandbox because ping failed '
                'with error: %s.',
                self.id,
                str(e)
            )
            self._housekeep_counter += 1
            self.report_state_error(e)
            self.on_housekeep(time.time() - housekeep_start, e)
            self.shutdown()
            break
          last_ping = time.time()

      for name, feature in self._features.items():
        if feature.housekeep_interval is not None and (
            time.time() - last_housekeep_time[name]
            > feature.housekeep_interval
        ):
          try:
            feature.housekeep()
            last_housekeep_time[name] = time.time()
          except interface.SandboxStateError as e:
            pg.logging.error(
                '[%s/%s]: Feature housekeeping failed with error: %s. '
                'Shutting down sandbox...',
                self.id,
                feature.name,
                e,
            )
            self.report_state_error(e)
            self._housekeep_counter += 1
            self.on_housekeep(time.time() - housekeep_start, e)
            self.shutdown()
            break

      self._housekeep_counter += 1
      self.on_housekeep(time.time() - housekeep_start)
      time.sleep(_next_housekeep_wait_time())

  #
  # Event handlers subclasses can override.
  #

  def on_start(
      self,
      duration: float,
      error: BaseException | None = None
  ) -> None:
    """Called when the sandbox is started."""
    self._event_handler.on_sandbox_start(self, duration, error)

  def on_status_change(
      self,
      old_status: interface.Sandbox.Status,
      new_status: interface.Sandbox.Status,
  ) -> None:
    """Called when the sandbox status changes."""
    status_duration = time.time() - self._status_start_time
    self._event_handler.on_sandbox_status_change(
        self, old_status, new_status, status_duration
    )

  def on_shutdown(
      self,
      duration: float,
      error: BaseException | None = None
  ) -> None:
    """Called when the sandbox is shutdown."""
    self._event_handler.on_sandbox_shutdown(
        sandbox=self,
        duration=duration,
        lifetime=(0.0 if self._start_time is None
                  else (time.time() - self._start_time)),
        error=error
    )

  def on_housekeep(
      self,
      duration: float,
      error: BaseException | None = None,
      **kwargs
  ) -> None:
    """Called when the sandbox finishes a round of housekeeping."""
    self._event_handler.on_sandbox_housekeep(
        sandbox=self,
        counter=self._housekeep_counter,
        duration=duration,
        error=error,
        **kwargs
    )

  def on_session_start(
      self,
      session_id: str,
      duration: float,
      error: BaseException | None = None
  ) -> None:
    """Called when the user session starts."""
    self._event_handler.on_sandbox_session_start(
        sandbox=self,
        session_id=session_id,
        duration=duration,
        error=error
    )

  def on_activity(
      self,
      name: str,
      duration: float,
      error: BaseException | None = None,
      **kwargs
  ) -> None:
    """Called when a sandbox activity is performed."""
    self._event_handler.on_sandbox_activity(
        name=name,
        sandbox=self,
        session_id=self.session_id,
        duration=duration,
        error=error,
        **kwargs
    )

  def on_session_end(
      self,
      session_id: str,
      duration: float,
      error: BaseException | None = None
  ) -> None:
    """Called when the user session ends."""
    self._event_handler.on_sandbox_session_end(
        sandbox=self,
        session_id=session_id,
        duration=duration,
        lifetime=time.time() - self._session_start_time,
        error=error
    )



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base for OpenAI compatible models (including OpenAI)."""

from typing import Annotated, Any

import langfun.core as lf
from langfun.core import modalities as lf_modalities
from langfun.core.data.conversion import openai as openai_conversion  # pylint: disable=unused-import
from langfun.core.llms import rest
import pyglove as pg


@lf.use_init_args(['api_endpoint', 'model'])
class OpenAIChatCompletionAPI(rest.REST):
  """Base class for models compatible with OpenAI's Chat Completion API.

  This class provides a common interface for language models that adhere to
  the OpenAI Chat Completion API format, which is used by providers like
  Groq, DeepSeek, and others. It standardizes request formatting and
  response parsing for these models.

  **References:**

  *   https://platform.openai.com/docs/api-reference/chat
  """

  model: Annotated[
      str, 'The name of the model to use.',
  ] = ''

  @property
  def headers(self) -> dict[str, Any]:
    return {
        'Content-Type': 'application/json'
    }

  def _request_args(
      self, options: lf.LMSamplingOptions) -> dict[str, Any]:
    """Returns a dict as request arguments."""
    # Reference:
    # https://platform.openai.com/docs/api-reference/completions/create
    # NOTE(daiyip): options.top_k is not applicable.
    args = {}

    if self.model:
      args['model'] = self.model
    if options.n != 1:
      args['n'] = options.n
    if options.top_logprobs is not None:
      args['top_logprobs'] = options.top_logprobs
    if options.logprobs:
      args['logprobs'] = options.logprobs
    if options.temperature is not None:
      args['temperature'] = options.temperature
    if options.max_tokens is not None:
      args['max_completion_tokens'] = options.max_tokens
    if options.top_p is not None:
      args['top_p'] = options.top_p
    if options.stop:
      args['stop'] = options.stop
    if options.random_seed is not None:
      args['seed'] = options.random_seed
    if options.reasoning_effort is not None:
      args['reasoning_effort'] = options.reasoning_effort
    if options.extras:
      args.update(options.extras)
    return args

  def request(
      self,
      prompt: lf.Message,
      sampling_options: lf.LMSamplingOptions
  ) -> dict[str, Any]:
    """Returns the JSON input for a message."""
    request_args = self._request_args(sampling_options)

    # Handle structured output.
    output_schema = self._structure_output_schema(prompt)
    if output_schema is not None:
      request_args.update(
          response_format=dict(
              type='json_schema',
              json_schema=output_schema,
          )
      )
      prompt.metadata.formatted_text = (
          prompt.text
          + '\n\n [RESPONSE FORMAT (not part of prompt)]\n'
          + pg.to_json_str(request_args['response_format'], json_indent=2)
      )

    # Prepare messages.
    messages = []

    def modality_check(chunk: str | lf.Modality) -> Any:
      if (isinstance(chunk, lf_modalities.Mime)
          and not self.supports_input(chunk.mime_type)):
        raise ValueError(
            f'Unsupported modality: {chunk!r}.'
        )
      return chunk

    # Users could use `metadata_system_message` to pass system message.
    system_message = prompt.metadata.get('system_message')
    if system_message:
      assert isinstance(system_message, lf.SystemMessage), type(system_message)
      messages.append(
          system_message.as_format(
              'openai_chat_completion_api', chunk_preprocessor=modality_check
          )
      )
    messages.append(
        prompt.as_format(
            'openai_chat_completion_api',
            chunk_preprocessor=modality_check
        )
    )
    request = dict()
    request.update(request_args)
    request['messages'] = messages
    return request

  def _structure_output_schema(
      self, prompt: lf.Message
  ) -> dict[str, Any] | None:
    # Users could use `metadata_json_schema` to pass additional
    # request arguments.
    json_schema = prompt.metadata.get('json_schema')
    if json_schema is not None:
      if not isinstance(json_schema, dict):
        raise ValueError(
            f'`json_schema` must be a dict, got {json_schema!r}.'
        )
      if 'title' not in json_schema:
        raise ValueError(
            f'The root of `json_schema` must have a `title` field, '
            f'got {json_schema!r}.'
        )
      return dict(
          schema=json_schema,
          name=json_schema['title'],
          strict=True,
      )
    return None

  def _parse_choice(self, choice: dict[str, Any]) -> lf.LMSample:
    # Reference:
    # https://platform.openai.com/docs/api-reference/chat/object
    logprobs = None
    choice_logprobs = choice.get('logprobs')
    if choice_logprobs:
      logprobs = [
          (
              t['token'],
              t['logprob'],
              [(tt['token'], tt['logprob']) for tt in t['top_logprobs']],
          )
          for t in choice_logprobs['content']
      ]
    return lf.LMSample(
        lf.Message.from_value(
            choice['message'],
            format='openai_chat_completion_api'
        ),
        score=0.0,
        logprobs=logprobs,
    )

  def result(self, json: dict[str, Any]) -> lf.LMSamplingResult:
    """Returns a LMSamplingResult from a JSON response."""
    usage = json['usage']
    return lf.LMSamplingResult(
        samples=[self._parse_choice(choice) for choice in json['choices']],
        usage=lf.LMSamplingUsage(
            prompt_tokens=usage['prompt_tokens'],
            completion_tokens=usage['completion_tokens'],
            total_tokens=usage['total_tokens'],
            completion_tokens_details=usage.get(
                'completion_tokens_details', None
            ),
        ),
    )

  def _error(self, status_code: int, content: str) -> lf.LMError:
    if (status_code == 413
        or (status_code == 400 and b'string_above_max_length' in content)):
      return lf.ContextLimitError(f'{status_code}: {content}')
    return super()._error(status_code, content)


class OpenAIResponsesAPI(OpenAIChatCompletionAPI):
  """Base class for models compatible with OpenAI's Responses API.

  This class provides a common interface for language models that adhere to
  the new OpenAI Responses API format. It standardizes request formatting
  and response parsing for these models, including handling instructions
  (system messages) and structured outputs.

  **References:**

  *   https://platform.openai.com/docs/api-reference/responses
  """

  def _request_args(
      self, options: lf.LMSamplingOptions) -> dict[str, Any]:
    """Returns a dict as request arguments."""
    if options.logprobs:
      raise ValueError('logprobs is not supported on Responses API.')
    if options.n != 1:
      raise ValueError('n must be 1 for Responses API.')
    return super()._request_args(options)

  def request(
      self,
      prompt: lf.Message,
      sampling_options: lf.LMSamplingOptions
  ) -> dict[str, Any]:
    """Returns the JSON input for a message."""
    request_args = self._request_args(sampling_options)

    # Handle structured output.
    output_schema = self._structure_output_schema(prompt)
    if output_schema is not None:
      output_schema['type'] = 'json_schema'
      request_args.update(text=dict(format=output_schema))
      prompt.metadata.formatted_text = (
          prompt.text
          + '\n\n [RESPONSE FORMAT (not part of prompt)]\n'
          + pg.to_json_str(request_args['text'], json_indent=2)
      )

    request = dict()
    request.update(request_args)

    # Users could use `metadata_system_message` to pass system message.
    system_message = prompt.metadata.get('system_message')
    if system_message:
      assert isinstance(system_message, lf.SystemMessage), type(system_message)
      request['instructions'] = system_message.text

    # Prepare input.
    def modality_check(chunk: str | lf.Modality) -> Any:
      if (isinstance(chunk, lf_modalities.Mime)
          and not self.supports_input(chunk.mime_type)):
        raise ValueError(
            f'Unsupported modality: {chunk!r}.'
        )
      return chunk

    request['input'] = [
        prompt.as_format(
            'openai_responses_api',
            chunk_preprocessor=modality_check
        )
    ]
    return request

  def _parse_output(self, output: dict[str, Any]) -> lf.LMSample:
    for item in output:
      if isinstance(item, dict) and item.get('type') == 'message':
        return lf.LMSample(
            lf.Message.from_value(item, format='openai_responses_api'),
            score=0.0,
        )
    raise ValueError('No message found in output.')

  def result(self, json: dict[str, Any]) -> lf.LMSamplingResult:
    """Returns a LMSamplingResult from a JSON response."""
    usage = json['usage']
    return lf.LMSamplingResult(
        samples=[self._parse_output(json['output'])],
        usage=lf.LMSamplingUsage(
            prompt_tokens=usage['input_tokens'],
            completion_tokens=usage['output_tokens'],
            total_tokens=usage['total_tokens'],
            completion_tokens_details=usage.get(
                'output_tokens_details', None
            ),
        ),
    )



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python sandboxing."""

import abc
import os
import tempfile
from typing import Annotated

from langfun.core.coding.python import parsing
import pyglove as pg


class SandboxOutput(pg.Object):
  """A structure containing the output from a sandbox execution.

  Attributes:
    stdout: The standard output captured during execution.
    stderr: The standard error captured during execution.
    output_files: A dictionary of file names to their byte content for files
      generated during execution.
  """

  stdout: Annotated[
      str,
      'The stdout of the sandbox execution.'
  ] = ''

  stderr: Annotated[
      str,
      'The stderr of the sandbox execution.'
  ] = ''

  output_files: Annotated[
      dict[str, bytes],
      'The output files of the sandbox execution.'
  ] = {}


class BaseSandbox(pg.Object):
  """Base class for Python code sandboxing.

  A sandbox provides an isolated environment for executing Python code,
  typically with restrictions on file system access, network calls, or other
  potentially harmful operations. This base class defines the interface for
  sandboxes, including methods for running code (`run`), uploading files
  (`upload`), and managing the sandbox lifecycle (`setup`, `cleanup`).
  """

  def _on_bound(self):
    super()._on_bound()
    self._uploaded_files: dict[str, str] = {}

  def run(
      self,
      code: str,
      *,
      timeout: int | None = 30,
      **kwargs
  ) -> SandboxOutput:
    """Runs code in the sandbox. Raises pg.coding.CodeError if failed."""
    return self._run(self.normalize_code(code), timeout=timeout, **kwargs)

  def normalize_code(self, code: str) -> str:
    """Returns normalized code runnable in the sandbox."""
    for original_path, uploaded_path in self._uploaded_files.items():
      code = code.replace(original_path, uploaded_path)
    return parsing.clean(code)

  @abc.abstractmethod
  def _run(
      self,
      code: str,
      *,
      timeout: int | None = 30,
      **kwargs
  ) -> SandboxOutput:
    """Runs code in the sandbox. Raises pg.coding.CodeError if failed."""

  def upload(self, path: str) -> str:
    """Uploads a file to the sandbox. Returns the uploaded path."""
    uploaded_path = self._upload(path)
    self._uploaded_files[path] = uploaded_path
    return uploaded_path

  @abc.abstractmethod
  def _upload(self, path: str) -> str:
    """Uploads a file to the sandbox."""

  def setup(self) -> None:
    """Sets up the sandbox."""
    self._uploaded_files = {}
    self._setup()

  def _setup(self) -> None:
    """Sets up the sandbox."""

  def cleanup(self) -> None:
    """Cleans up the sandbox."""
    self._uploaded_files = {}
    self._cleanup()

  def _cleanup(self) -> None:
    """Cleans up the sandbox."""

  def __enter__(self):
    """Enters the sandbox."""
    self.setup()
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    """Exits the sandbox."""
    self.cleanup()


class MultiProcessingSandbox(BaseSandbox):
  """A sandbox implementation using Python's `multiprocessing`.

  This sandbox executes code in a separate process, providing isolation from
  the main process. It uses a temporary directory for file operations,
  which is cleaned up when the sandbox is closed. It relies on
  `pg.coding.run` with `sandbox=True` for execution.
  """

  def _on_bound(self):
    super()._on_bound()
    self._working_dir = None

  @property
  def working_dir(self) -> str | None:
    """Returns the directory of the sandbox."""
    return self._working_dir

  def _setup(self) -> None:
    """Sets up the sandbox."""
    self._working_dir = tempfile.TemporaryDirectory()

  def _cleanup(self) -> None:
    """Cleans up the sandbox."""
    assert self._working_dir is not None
    self._working_dir.cleanup()

  def _run(
      self,
      code: str,
      *,
      timeout: int | None = 30,
      **kwargs
  ) -> SandboxOutput:
    """Runs code in the sandbox."""
    stdout = pg.coding.run(
        code, returns_stdout=True, sandbox=True, timeout=timeout
    )
    return SandboxOutput(stdout=stdout)

  def _upload(self, path: str) -> str:
    """Uploads a file to the sandbox."""
    if self._working_dir is None:
      raise ValueError('Sandbox is not set up.')

    # Upload the file to the sandbox directory.
    uploaded_path = os.path.join(self._working_dir.name, os.path.basename(path))
    with pg.io.open(path, 'r') as r:
      with pg.io.open(uploaded_path, 'w') as w:
        w.write(r.read())
    return uploaded_path



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common value types for evaluation metrics and metadata."""


import abc
from typing import Annotated, Any, Union
import pyglove as pg


class MetricValue(pg.Object):
  """Base class for metric values.

  `MetricValue` is the base class for representing aggregated metric values
  in an evaluation. It accumulates data points from individual examples,
  each consisting of a value and an optional weight, associated with an example
  ID. Subclasses must implement `reduce` method to compute a single float value
  from accumulated data points, and `scalar_repr` to provide a string
  representation of the reduced value.
  """

  class DataPoint(pg.Object):
    """A data point for a metric value."""
    example_id: int
    value: float
    weight: float = 1.0

  # NOTE(daiyip): For evaluations, usually the number of examples is within 10K,
  # therefore it's beneficial to store all accumulated values with their example
  # IDs so we are able to track the individual examples that contributed to this
  # metric value. If this premise changes, we might consider using a more
  # efficient data structure.
  data_points: Annotated[
      list[DataPoint],
      'Accumulated computed values with example IDs and weights.'
  ] = []

  total: Annotated[
      int,
      'The total number of examples being evaluated. Including errors.'
  ] = 0

  def _on_bound(self):
    super()._on_bound()
    self._weighted_sum = sum(dp.value * dp.weight for dp in self.data_points)

  def reset(self) -> None:
    """Resets the value to its initial state."""
    self._sync_members(data_points=[], total=0)
    self._weighted_sum = 0.0

  def _sync_members(self, **kwargs) -> None:
    """Synchronizes the members of this object."""
    self.rebind(**kwargs, skip_notification=True, raise_on_no_change=False)

  def __float__(self) -> float:
    """Returns the float representation of this object."""
    if self.total == 0:
      return float('nan')
    return self.reduce()

  @abc.abstractmethod
  def reduce(self) -> float:
    """Reduces the accumulated values into a single value."""

  def increment_total(self, delta: int = 1) -> 'MetricValue':
    """Increments the total number of examples being evaluated."""
    self._sync_members(total=self.total + delta)
    return self

  def add(
      self,
      example_id: int,
      value: float,
      weight: float = 1.0,
      increment_total: bool = False,
  ) -> 'MetricValue':
    """Adds a value to the accumulated values."""
    self._weighted_sum += value * weight
    with pg.notify_on_change(False), pg.allow_writable_accessors(True):
      self.data_points.append(
          MetricValue.DataPoint(example_id, value, weight)
      )
      if increment_total:
        self.increment_total()
    return self

  def merge_from(self, other: 'MetricValue') -> 'MetricValue':
    """Merges the values from another metric value."""
    self._weighted_sum += other._weighted_sum  # pylint: disable=protected-access
    with pg.notify_on_change(False), pg.allow_writable_accessors(True):
      self.data_points.extend(other.data_points)
      self.increment_total(other.total)
    return self

  def __gt__(self, other: Union['MetricValue', float]) -> bool:
    if isinstance(other, self.__class__):
      return float(self) > float(other)
    return float(self) > other

  def __lt__(self, other: Union['MetricValue', float]) -> bool:
    if isinstance(other, self.__class__):
      return float(self) < float(other)
    return float(self) < other

  def __eq__(self, other: Union['MetricValue', float]) -> bool:
    if isinstance(other, self.__class__):
      return super().__eq__(other)
    return float(self) == other

  def __nonzero__(self) -> bool:
    return float(self) != 0

  def format(
      self,
      compact: bool = False,
      verbose: bool = True,
      *args,
      **kwargs
  ) -> str:
    if compact:
      return super().format(compact, *args, **kwargs)
    if self.total == 0:
      return 'n/a'
    if verbose:
      return (
          f'{self.scalar_repr()} ({len(self.data_points)}/{self.total})'
      )
    return self.scalar_repr()

  @abc.abstractmethod
  def scalar_repr(self) -> str:
    """Returns the format string for the value."""

  def _sym_nondefault(self) -> dict[str, Any]:
    """Overrides nondefault valuesso volatile values are not included."""
    return dict()


class Rate(MetricValue):
  """Metric value representing a rate in range [0, 1].

  `Rate` is used for metrics that compute a rate, such as accuracy or error
  rate. The final value is computed as the weighted sum of accumulated values
  divided by the total number of examples. It's displayed as a percentage
  (e.g., 90.0%).
  """

  def reduce(self) -> float:
    return self._weighted_sum / self.total

  def scalar_repr(self):
    if self.total == 0:
      return 'n/a'
    return f'{self.reduce():.1%}'


class Average(MetricValue):
  """Metric value representing an average of accumulated values.

  `Average` is used for metrics that compute an average score across examples
  (e.g., average quality score). The final value is computed as the weighted
  sum of accumulated values divided by the number of data points.
  It's displayed as a float with 3 decimal places (e.g., 4.750).
  """

  def reduce(self) -> float:
    if not self.data_points:
      return float('nan')
    return self._weighted_sum / len(self.data_points)

  def scalar_repr(self):
    return f'{self.reduce():.3f}'



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""LangFunc: Language-based functions."""

import dataclasses
from typing import Annotated, Type

from langfun.core import component
from langfun.core import language_model
from langfun.core import message as message_lib
from langfun.core import subscription
from langfun.core import template as template_lib
import pyglove as pg


# Keys for storing render/call stack (invoked LangFunc objects) in the
# thread-local storage.
_TLS_LFUN_CALL_STACK = '_langfunc_callstack'


# NOTE(daiyip): Only the template string belongs to the positional arguments,
# all others are keyword-only for clarity.
@pg.use_init_args(['template_str'])
class LangFunc(template_lib.Template):
  r"""Base class for Language-based functions.

  LangFunc represents a function powered by a language model. It is a subclass
  of `lf.Template` and can be thought of as a `lf.Template` augmented with an LM
  and an output transformation. Calling a `lf.LangFunc` is equivalent to calling
  the LM with the rendered prompt and transforming the output.

  LangFunc can be directly constructed and used.

  ```python
  import langfun as lf

  func = lf.LangFunc("Hello, {{name}}!")
  print(func(name="Gemini", lm=lf.llms.Gemini25Flash()))
  # Output: Hello, how are you today?
  ```

  Or it can be subclassed:

  ```python
  import langfun as lf

  class Compute(lf.LangFunc):
    '''Compute a simple arithmetic expression.

    {{expression}} = ?
    '''
    expression: str

    def transform_output(self, lm_output: lf.Message) -> lf.Message:
      lm_output.metadata.result = float(lm_output.text)
      return lm_output

  r = Compute(expression="1 + 1")(lm=lf.llms.Gemini25Flash())
  print(r.result)
  # Output: 2.0

  Final note: always include these capitalized words if you don't want to treat
  the docstr as the template str: THIS IS NOT A TEMPLATE. So as a result, this
  docstr is not treated as a template str :).
  """

  lm: Annotated[
      language_model.LanguageModel,
      (
          'The language model used to sample the responses based on the '
          'rendered template. It could either be provided at `__init__` time '
          'or accessed at `__call__` time from its context.'
      ),
  ] = component.contextual()

  def _on_bound(self):
    super()._on_bound()

    # Last LM input and output.
    self._cached_lm_input = None
    self._cached_lm_output = None

  @property
  def lm_input(self) -> message_lib.Message | None:
    """Returns the cached LM input from the last invocation to `__call__`."""
    return self._cached_lm_input

  @property
  def lm_output(self) -> message_lib.Message | None:
    """Returns the cached LM output from the last invocation to `__call__`."""
    return self._cached_lm_output

  def __call__(
      self,
      *,
      lm: language_model.LanguageModel | None = None,
      lm_input: message_lib.Message | None = None,
      cache_seed: int | None = 0,
      skip_lm: bool = False,
      **variables,
  ) -> message_lib.Message:
    """Calls language model with `lm_input` or rendered text.

    Args:
      lm: Language model to use. When present, it takes precedence over the
        value from the `lm` attribute as well as from the containing chain.
      lm_input: An optional `lf.Message` object as the `lm_input`. If present,
        it will be directly used to call the LM. Otherwise, `render()` will be
        called to produce an LM input first.
      cache_seed: Seed for computing cache key. The cache key is determined by a
        tuple of (lm, prompt, cache seed). If None, cache will be disabled for
        the query even cache is configured by the LM.
      skip_lm: If True, returns the rendered prompt as a UserMessage object.
        otherwise return the LLM response based on the rendered prompt.
      **variables: Template variables applicable to this or child LangFunc.

    Returns:
      An output message from LM.
    """
    return self._call_once(
        lm=lm,
        lm_input=lm_input,
        cache_seed=cache_seed,
        skip_lm=skip_lm,
        **variables,
    )

  def _call_once(
      self,
      *,
      lm: language_model.LanguageModel | None = None,
      lm_input: message_lib.Message | None = None,
      cache_seed: int | None = 0,
      skip_lm: bool = False,
      **variables,
  ) -> message_lib.Message:
    """Call the language model once, with invoking the output transform."""
    try:
      pg.object_utils.thread_local_push(_TLS_LFUN_CALL_STACK, self)

      kwargs = dict(variables)
      if lm is not None:
        kwargs['lm'] = lm

      with self.override(**kwargs):
        # Render the LM input text and creates a user message.
        if lm_input is None:
          lm_input = self.render(**kwargs)

        if skip_lm:
          return lm_input

        self._cached_lm_input = lm_input

        # Send rendered text to LM.
        lm_output = self.lm(lm_input, cache_seed=cache_seed)

        # Attach cache seed.
        lm_input.metadata.cache_seed = cache_seed

        # Transform the output message.
        lm_output = self.transform_output(lm_output)
        lm_output.tag(message_lib.Message.TAG_LM_OUTPUT)

        # We cache the transformed output instead of the original one
        # since the old one is tracked with `sym_origin`.
        self._cached_lm_output = lm_output

      # Emit LangFuncCallEvent.
      lm_callstack = list(
          pg.object_utils.thread_local_get(_TLS_LFUN_CALL_STACK))
      lm_callstack.pop()
      subscription.emit(
          LangFuncCallEvent(
              sender=self,
              lm_input=lm_input,
              lm_output=lm_output,
              lm_callstack=lm_callstack,
          )
      )
      return lm_output
    finally:
      top = pg.object_utils.thread_local_pop(_TLS_LFUN_CALL_STACK, self)
      assert top is self, (top, self)

  def render(
      self,
      *,
      allow_partial: bool = False,
      implicit: bool = False,
      message_cls: Type[message_lib.Message] = message_lib.UserMessage,
      **kwargs,
  ) -> message_lib.Message:
    """Renders the template and transforms it as LM input message.

    Args:
      allow_partial: If True, allows partial rendering, which leaves unresolved
        variables in place in the output text. Otherwise, raises error when
        there are unresolved variables.
      implicit: If True, reuse the rendering output if a parent `lf.Template`
        is rendering current `lf.Template` multiple times. This is important
        for making sure all references to the same `lf.Template` within a single
        top-level rendering would return the same result. If False, every call
        to `render` will trigger the actual rendering process.
      message_cls: The message class used for creating the return value.
      **kwargs: Values for template variables, which override values from
        member attributes or context.

    Returns:
      A Message object containing the rendered result.
    """
    lm_input = super().render(
        allow_partial=allow_partial,
        implicit=implicit,
        message_cls=message_cls,
        **kwargs,
    )

    with component.context(**kwargs):
      with self.override(**kwargs):
        return self.transform_input(lm_input)

  #
  # Input and output transforms.
  # Subclasses can override.
  #

  def transform_input(
      self, lm_input: message_lib.Message) -> message_lib.Message:
    """Transforms the input message before sending to LM."""
    return lm_input

  def transform_output(
      self, lm_output: message_lib.Message) -> message_lib.Message:
    """Transforms the output message before returning from __call__."""
    return lm_output


# Register converter from str to LangFunc, therefore we can always
# pass strs to attributes that accept LangFunc.
pg.typing.register_converter(str, LangFunc, LangFunc)


#
# LangFunc events.
#


@dataclasses.dataclass
class LangFuncCallEvent(subscription.Event[LangFunc]):
  """LangFunc call event."""

  lm_input: message_lib.Message
  lm_output: message_lib.Message
  lm_callstack: list[LangFunc]



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Config saver plugins."""

import os
from langfun.core.eval.v2 import experiment as experiment_lib


class RunConfigSaver(experiment_lib.Plugin):
  """Saves the current run."""

  def on_run_start(
      self,
      runner: experiment_lib.Runner,
      root: experiment_lib.Experiment
  ) -> None:
    del root  # Unused.
    self._save_run_config(runner)

  def _save_run_config(self, runner: experiment_lib.Runner) -> None:
    def _save():
      runner.current_run.save(
          os.path.join(runner.current_run.output_root, 'run.json'),
          hide_default_values=True,
      )
    runner.background_run(_save)



"""Langfun MCP support."""

# pylint: disable=g-importing-member

from langfun.core.mcp.client import McpClient
from langfun.core.mcp.session import McpSession
from langfun.core.mcp.tool import McpTool
from langfun.core.mcp.tool import McpToolInput

# pylint: enable=g-importing-member



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parallel runner."""

import collections
import math
import random
import threading
import time

from typing import Annotated, Iterator
import langfun.core as lf
from langfun.core.eval.v2 import checkpointing
from langfun.core.eval.v2 import experiment as experiment_lib
from langfun.core.eval.v2.runners import base
from langfun.core.eval.v2.runners import ckpt_monitor
import pyglove as pg


class ParallelRunner(base.RunnerBase):
  """A runner that executes evaluations and examples in parallel.

  The parallel runner groups evaluations by their required resources
  (e.g., specific LLMs) and runs evaluations that do not share resources in
  parallel. Within each evaluation, examples are also processed in parallel
  using threads, up to `Evaluation.max_workers`.
  """

  NAME = 'parallel'

  timeout: Annotated[
      int | None,
      'Timeout for each evaluation example.'
  ] = None

  concurrent_startup_delay: Annotated[
      tuple[int, int] | None,
      (
          'A range of seconds to delay the initial evaluation of each thread '
          'in the thread pool, helping to prevent a burst in LLM QPS at '
          'startup. If set to None, no delay will be applied.'
      )
  ] = None

  def _run(self, evaluations: list[base.Evaluation]) -> None:
    """Runs the evaluations in parallel."""
    def _run_group(evaluation_group: list[base.Evaluation]):
      for e in evaluation_group:
        self.run_evaluation(e)

    # Run evaluations in parallel groupped by resource key.
    groups: dict[str, list[base.Evaluation]] = collections.defaultdict(list)
    for e in evaluations:
      resource_ids = e.resource_ids()
      if not resource_ids:
        group_id = e.id
      else:
        # TODO(daiyip): support group that requires multiple resources.
        group_id = resource_ids.pop()
      groups[group_id].append(e)

    for _, _, _ in lf.concurrent_map(
        _run_group,
        groups.values(),
        max_workers=max(64, len(groups)),
        timeout=self.timeout,
        silence_on_errors=None,
    ):
      pass

  def _evaluate_items(
      self, evaluation: base.Evaluation, items: Iterator[base.Example]
  ) -> None:
    """Override run items to run in parallel."""
    if self.concurrent_startup_delay is not None:
      thread_delayed = {}
      def _evaluate_item(item: base.Example):
        thread_id = threading.current_thread().ident
        if thread_id not in thread_delayed:
          thread_delayed[thread_id] = True
          time.sleep(random.randint(*self.concurrent_startup_delay))
        return self.evaluate_item(evaluation, item)
    else:
      def _evaluate_item(item: base.Example):
        return self.evaluate_item(evaluation, item)

    for _, _, _ in lf.concurrent_map(
        _evaluate_item,
        items,
        max_workers=self._max_workers(evaluation),
        timeout=self.timeout,
        silence_on_errors=None,
    ):
      pass

  def _max_workers(self, evaluation: base.Evaluation) -> int | None:
    return evaluation.max_workers


class _SingleSliceRunner(ParallelRunner):
  """A single slice runner."""

  NAME = '__single_slice_runner__'

  # Do not track progress in single slice runner.
  progress_tracker = None

  num_slices: Annotated[
      int,
      'The number of slices to run the evaluations in.'
  ] = 1

  def _max_workers(self, evaluation: base.Evaluation) -> int | None:
    max_workers = super()._max_workers(evaluation)
    if max_workers is None:
      return None
    return max(1, math.ceil(max_workers / self.num_slices))


class MultiSliceParallelRunner(experiment_lib.Runner):
  """A sliced parallel runner.

  An evaluation is split into `num_slices` slices. Each MultiSliceParallelRunner
  instance is responsible for evaluating a single slice. The instance with
  `slice_id` 0 will also aggregate checkpoints from all slices.

  Sliced parallel runner allows running multiple instances across different
  machines/hosts parallelly. This can be utilize for scaling the evaluation
  jobs to run on multiple machines, or running evaluations in a fault-tolerant
  way by splitting each evaluation into multiple slices.
  """

  NAME = 'sliced-parallel'

  slice_id: Annotated[
      int,
      (
          'The slice ID of the runner. If 0, it will also run as the '
          'aggregator for collecting results from other slices. '
      )
  ] = 0

  num_slices: Annotated[
      int,
      'The number of slices to run the evaluations in parallel.'
  ] = 1

  timeout: Annotated[
      int | None,
      'Timeout for each evaluation example.'
  ] = None

  concurrent_startup_delay: Annotated[
      tuple[int, int] | None,
      (
          'A range of seconds to delay the initial evaluation of each thread '
          'in the thread pool, helping to prevent a burst in LLM QPS at '
          'startup. If set to None, no delay will be applied.'
      )
  ] = None

  ckpt_format: Annotated[
      str,
      'The file extension of the checkpoint files.'
  ] = 'bagz'

  max_aggregation_threads: Annotated[
      int,
      'The maximum number of threads to aggregate checkpoints.'
  ] = 32

  def _on_bound(self):
    super()._on_bound()
    if self.current_run.use_cache != 'no':
      raise ValueError(
          'Cache is not supported in MultiProcessParallelRunner. '
          f'Encountered: {self.current_run.use_cache}'
      )
    monitor_plugins = []
    worker_plugins = [
        checkpointing.PerExampleCheckpointer(
            checkpoint_filename=f'checkpoint.{self.ckpt_format}'
        ),
    ]
    for plugin in self.plugins:
      if isinstance(plugin, checkpointing.Checkpointer):
        pg.logging.warning(
            'Built-in checkpointing is enabled on MultiProcessParallelRunner. '
            f'Ignoring checkpointer: {plugin!r}.'
        )
      elif plugin.is_per_example():
        worker_plugins.append(pg.Ref(plugin))
      else:
        monitor_plugins.append(pg.Ref(plugin))

    if self.slice_id == 0:
      self._ckpt_monitor = ckpt_monitor.CheckpointMonitor(
          pg.Ref(self.current_run),
          plugins=monitor_plugins,
          monitor_inprogress_files=True,
          checkpoint_pattern=f'checkpoint_*.{self.ckpt_format}',
          max_aggregation_threads=self.max_aggregation_threads,
      )
    else:
      self._ckpt_monitor = None

    self._slice_runner = _SingleSliceRunner(
        current_run=self.current_run.clone(
            override=dict(
                # Clone the experiment to avoid updating the original one.
                experiment=self.current_run.experiment.clone(),
                example_ids=self._examples_to_evaluate,
            )
        ),
        plugins=worker_plugins,
        timeout=self.timeout,
        concurrent_startup_delay=self.concurrent_startup_delay,
    )

  def _examples_to_evaluate(
      self,
      experiment: experiment_lib.Experiment
  ) -> list[int]:
    all_ids = self.current_run.examples_to_evaluate(experiment)
    return [x for x in all_ids if x % self.num_slices == self.slice_id]

  def run(self) -> None:
    if self._ckpt_monitor is not None:
      self._ckpt_monitor.start()
    self._slice_runner.run()
    if self._ckpt_monitor is not None:
      self._ckpt_monitor.join()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""OpenAI API message conversion."""

from typing import Annotated, Any, Callable

import langfun.core as lf
from langfun.core import modalities as lf_modalities


class OpenAIChatCompletionAPIMessageConverter(lf.MessageConverter):
  """Converter for OpenAI Chat Completion API.

  This converter translates `lf.Message` objects into the JSON format
  required by the OpenAI Chat Completions API
  (https://platform.openai.com/docs/api-reference/chat) and vice versa.
  It handles text and image modalities, mapping Langfun roles to OpenAI
  roles ('system', 'user', 'assistant'). An optional `chunk_preprocessor`
  can be provided to modify or filter chunks before conversion.
  """

  FORMAT_ID = 'openai_chat_completion_api'

  chunk_preprocessor: Annotated[
      Callable[[str | lf.Modality], Any] | None,
      (
          'Chunk preprocessor for Langfun chunk to OpenAI chunk conversion. '
          'It will be applied before each Langfun chunk is converted. '
          'If it returns None, the chunk will be skipped.'
      )
  ] = None

  def to_value(self, message: lf.Message) -> dict[str, Any]:
    """Converts a Langfun message to OpenAI API."""
    parts = []
    for chunk in message.chunk():
      if self.chunk_preprocessor is not None:
        chunk = self.chunk_preprocessor(chunk)
        if chunk is None:
          continue
      parts.append(self.chunk_to_json(type(message), chunk))
    return dict(
        role=self.get_role(message),
        content=parts,
    )

  def chunk_to_json(
      self,
      message_cls: type[lf.Message],
      chunk: str | lf.Modality
  ) -> dict[str, Any]:
    """Converts a Langfun chunk to OpenAI chunk."""
    del message_cls
    if isinstance(chunk, str):
      return dict(type='text', text=chunk)
    elif isinstance(chunk, lf_modalities.Image):
      return dict(
          type='image_url', image_url=dict(url=chunk.embeddable_uri)
      )
    # TODO(daiyip): Support audio_input.
    else:
      raise ValueError(f'Unsupported content type: {chunk!r}.')

  def get_role(self, message: lf.Message) -> str:
    """Returns the role of the message."""
    if isinstance(message, lf.SystemMessage):
      return 'system'
    elif isinstance(message, lf.UserMessage):
      return 'user'
    elif isinstance(message, lf.AIMessage):
      return 'assistant'
    else:
      raise ValueError(f'Unsupported message type: {message!r}.')

  def get_message_cls(self, role: str) -> type[lf.Message]:
    """Returns the message class of the message."""
    match role:
      case 'system':
        return lf.SystemMessage
      case 'user':
        return lf.UserMessage
      case 'assistant':
        return lf.AIMessage
      case _:
        raise ValueError(f'Unsupported role: {role!r}.')

  def from_value(self, value: dict[str, Any]) -> lf.Message:
    """Returns a Langfun message from OpenAI message."""
    message_cls = self.get_message_cls(
        self._safe_read(value, 'role', default='assistant')
    )
    content = self._safe_read(value, 'content')
    if isinstance(content, str):
      return message_cls(content)

    assert isinstance(content, list)
    chunks = []
    for item in content:
      chunks.append(self.json_to_chunk(item))
    return message_cls.from_chunks(chunks)

  def json_to_chunk(self, json: dict[str, Any]) -> str | lf.Modality:
    """Returns a Langfun chunk from OpenAI chunk JSON."""
    t = self._safe_read(json, 'type')
    if t == 'text':
      return self._safe_read(json, 'text')
    elif t == 'image_url':
      return lf_modalities.Image.from_uri(
          self._safe_read(self._safe_read(json, 'image_url'), 'url')
      )
    else:
      raise ValueError(f'Unsupported content type: {json!r}.')


def _as_openai_chat_completion_api_format(
    self,
    chunk_preprocessor: Callable[[str | lf.Modality], Any] | None = None,
    **kwargs
) -> dict[str, Any]:
  """Returns an OpenAI format message."""
  return OpenAIChatCompletionAPIMessageConverter(
      chunk_preprocessor=chunk_preprocessor, **kwargs
  ).to_value(self)


@classmethod
def _from_openai_chat_completion_api_format(
    cls,
    openai_message: dict[str, Any],
    **kwargs
) -> lf.Message:
  """Creates a Langfun message from the OpenAI format message."""
  del cls
  return OpenAIChatCompletionAPIMessageConverter(
      **kwargs
  ).from_value(openai_message)

# Set shortcut methods in lf.Message.
lf.Message.as_openai_chat_completion_api_format = (
    _as_openai_chat_completion_api_format
)

lf.Message.from_openai_chat_completion_api_format = (
    _from_openai_chat_completion_api_format
)


#
# OpenAI Responses API message converter.
#


class OpenAIResponsesAPIMessageConverter(
    OpenAIChatCompletionAPIMessageConverter
):
  """Converter for OpenAI Responses API.

  This converter translates `lf.Message` objects into the JSON format
  required by the OpenAI Responses API
  (https://platform.openai.com/docs/api-reference/responses/create),
  which is used for human-in-the-loop rating, and vice versa.
  It extends `OpenAIChatCompletionAPIMessageConverter` but uses different
  type names for content chunks (e.g., 'input_text', 'output_image').
  """

  FORMAT_ID = 'openai_responses_api'

  def to_value(self, message: lf.Message) -> dict[str, Any]:
    """Converts a Langfun message to OpenAI API."""
    message_json = super().to_value(message)
    message_json['type'] = 'message'
    return message_json

  def chunk_to_json(
      self,
      message_cls: type[lf.Message],
      chunk: str | lf.Modality
  ) -> dict[str, Any]:
    """Converts a Langfun chunk to OpenAI chunk."""
    source = 'output' if issubclass(message_cls, lf.AIMessage) else 'input'

    if isinstance(chunk, str):
      return dict(type=f'{source}_text', text=chunk)
    elif isinstance(chunk, lf_modalities.Image):
      return dict(
          type=f'{source}_image', image_url=chunk.embeddable_uri
      )
    # TODO(daiyip): Support audio_input.
    else:
      raise ValueError(f'Unsupported content type: {chunk!r}.')

  def json_to_chunk(self, json: dict[str, Any]) -> str | lf.Modality:
    """Returns a Langfun chunk from OpenAI chunk JSON."""
    t = self._safe_read(json, 'type')
    if t in ('input_text', 'output_text'):
      return self._safe_read(json, 'text')
    elif t in ('input_image', 'output_image'):
      return lf_modalities.Image.from_uri(self._safe_read(json, 'image_url'))
    else:
      raise ValueError(f'Unsupported content type: {json!r}.')


def _as_openai_responses_api_format(
    self,
    chunk_preprocessor: Callable[[str | lf.Modality], Any] | None = None,
    **kwargs
) -> dict[str, Any]:
  """Returns an OpenAI format message."""
  return OpenAIResponsesAPIMessageConverter(
      chunk_preprocessor=chunk_preprocessor, **kwargs
  ).to_value(self)


@classmethod
def _from_openai_responses_api_format(
    cls,
    openai_message: dict[str, Any],
    **kwargs
) -> lf.Message:
  """Creates a Langfun message from the OpenAI format message."""
  del cls
  return OpenAIResponsesAPIMessageConverter(
      **kwargs
  ).from_value(openai_message)


# Set shortcut methods in lf.Message.
lf.Message.as_openai_responses_api_format = (
    _as_openai_responses_api_format
)

lf.Message.from_openai_responses_api_format = (
    _from_openai_responses_api_format
)



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Environment event handler chain."""

from typing import Sequence
from langfun.env import interface


class EventHandlerChain(interface.EventHandler):
  """Environment event handler chain."""

  def __init__(self, handlers: Sequence[interface.EventHandler]):
    super().__init__()
    self._handlers = list(handlers)

  def add(self, handler: interface.EventHandler):
    self._handlers.append(handler)

  def remove(self, handler: interface.EventHandler):
    self._handlers.remove(handler)

  def on_environment_starting(
      self,
      environment: interface.Environment,
  ) -> None:
    """Called when the environment is starting."""
    for handler in self._handlers:
      handler.on_environment_starting(environment)

  def on_environment_shutting_down(
      self,
      environment: interface.Environment,
      offline_duration: float,
  ) -> None:
    """Called when the environment is shutting down."""
    for handler in self._handlers:
      handler.on_environment_shutting_down(environment, offline_duration)

  def on_environment_start(
      self,
      environment: interface.Environment,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when the environment is started."""
    for handler in self._handlers:
      handler.on_environment_start(environment, duration, error)

  def on_environment_housekeep(
      self,
      environment: interface.Environment,
      counter: int,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when the environment is housekeeping."""
    for handler in self._handlers:
      handler.on_environment_housekeep(
          environment, counter, duration, error, **kwargs
      )

  def on_environment_shutdown(
      self,
      environment: interface.Environment,
      duration: float,
      lifetime: float,
      error: BaseException | None
  ) -> None:
    """Called when the environment is shutdown."""
    for handler in self._handlers:
      handler.on_environment_shutdown(environment, duration, lifetime, error)

  def on_sandbox_start(
      self,
      sandbox: interface.Sandbox,
      duration: float,
      error: BaseException | None
  ) -> None:
    for handler in self._handlers:
      handler.on_sandbox_start(sandbox, duration, error)

  def on_sandbox_status_change(
      self,
      sandbox: interface.Sandbox,
      old_status: interface.Sandbox.Status,
      new_status: interface.Sandbox.Status,
      span: float
  ) -> None:
    for handler in self._handlers:
      handler.on_sandbox_status_change(sandbox, old_status, new_status, span)

  def on_sandbox_shutdown(
      self,
      sandbox: interface.Sandbox,
      duration: float,
      lifetime: float,
      error: BaseException | None
  ) -> None:
    for handler in self._handlers:
      handler.on_sandbox_shutdown(sandbox, duration, lifetime, error)

  def on_sandbox_session_start(
      self,
      sandbox: interface.Sandbox,
      session_id: str,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox session starts."""
    for handler in self._handlers:
      handler.on_sandbox_session_start(sandbox, session_id, duration, error)

  def on_sandbox_session_end(
      self,
      sandbox: interface.Sandbox,
      session_id: str,
      duration: float,
      lifetime: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox session ends."""
    for handler in self._handlers:
      handler.on_sandbox_session_end(
          sandbox, session_id, duration, lifetime, error
      )

  def on_sandbox_activity(
      self,
      name: str,
      sandbox: interface.Sandbox,
      session_id: str | None,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when a sandbox activity is performed."""
    for handler in self._handlers:
      handler.on_sandbox_activity(
          name, sandbox, session_id, duration, error, **kwargs
      )

  def on_sandbox_housekeep(
      self,
      sandbox: interface.Sandbox,
      counter: int,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when a sandbox feature is housekeeping."""
    for handler in self._handlers:
      handler.on_sandbox_housekeep(sandbox, counter, duration, error, **kwargs)

  def on_feature_setup(
      self,
      feature: interface.Feature,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is setup."""
    for handler in self._handlers:
      handler.on_feature_setup(feature, duration, error)

  def on_feature_teardown(
      self,
      feature: interface.Feature,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is teardown."""
    for handler in self._handlers:
      handler.on_feature_teardown(feature, duration, error)

  def on_feature_setup_session(
      self,
      feature: interface.Feature,
      session_id: str | None,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is setup."""
    for handler in self._handlers:
      handler.on_feature_setup_session(feature, session_id, duration, error)

  def on_feature_teardown_session(
      self,
      feature: interface.Feature,
      session_id: str,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is teardown."""
    for handler in self._handlers:
      handler.on_feature_teardown_session(feature, session_id, duration, error)

  def on_feature_activity(
      self,
      name: str,
      feature: interface.Feature,
      session_id: str | None,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when a feature activity is performed."""
    for handler in self._handlers:
      handler.on_feature_activity(
          name, feature, session_id, duration, error, **kwargs
      )

  def on_feature_housekeep(
      self,
      feature: interface.Feature,
      counter: int,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when a sandbox feature is housekeeping."""
    for handler in self._handlers:
      handler.on_feature_housekeep(feature, counter, duration, error, **kwargs)



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base classes for Langfun evaluation."""

import dataclasses
import inspect
from typing import Any, Callable, Iterator
import langfun.core as lf
import pyglove as pg


@dataclasses.dataclass
class Example(pg.JSONConvertible, pg.views.HtmlTreeView.Extension):
  """An example for evaluation.

  An evaluation example contains the input and output of an evaluation task,
  as well as metadata about the evaluation process, such as execution time,
  LLM usage, and metric results.

  Attributes:
    id: The 1-based ID of the example in the evaluation set.
    input: An element returned from the `Evaluable.inputs` functor, which serves
      as the input for `lf.Evaluable.process`.
    output: The output of `lf.Evaluable.process` method. If `pg.MISSING_VALUE`,
      it indicates the example has not been processed yet.
    error: The error raised from `lf.Evaluable.process`. If None, it
      indicates the process was successful.
    metadata: The metadata of the example produced by `lf.Evaluable.process`.
    metric_metadata: The dictionary returned from `Metric.audit`, which contains
      metadata about metric computation for this example.
    newly_processed: Whether this example is processed in the current run. If
      False, it indicates the example was loaded from a checkpoint from previous
      runs.
    start_time: The start time of processing this example.
    end_time: The end time of processing this example.
    usage_summary: The summary of LLM usages for processing this example.
    execution_status: The timeit status of processing this example.
  """
  id: int
  input: Any = pg.MISSING_VALUE
  output: Any = pg.MISSING_VALUE
  error: pg.ErrorInfo | None = None
  metadata: dict[str, Any] = dataclasses.field(default_factory=dict)
  metric_metadata: dict[str, Any] | None = None
  # Execution information.
  newly_processed: bool = True
  start_time: float | None = None
  end_time: float | None = None
  usage_summary: lf.UsageSummary | None = None
  execution_status: dict[str, pg.utils.TimeIt.Status] | None = None

  @property
  def is_processed(self) -> bool:
    """Returns whether the item has been processed."""
    return pg.MISSING_VALUE != self.output

  @property
  def has_error(self) -> bool:
    """Returns whether the item has an error."""
    return self.error is not None

  @property
  def elapse(self) -> float | None:
    """Returns the elapse time of the item."""
    if self.execution_status is not None:
      return self.execution_status['evaluate'].elapse
    return None

  def to_json(self, *, exclude_input: bool = False, **kwargs):
    """Returns the JSON representation of the item."""
    return self.to_json_dict(
        fields=dict(
            id=(self.id, None),
            input=(
                self.input if not exclude_input else pg.MISSING_VALUE,
                pg.MISSING_VALUE
            ),
            output=(self.output, pg.MISSING_VALUE),
            error=(self.error, None),
            metadata=(self.metadata, {}),
            metric_metadata=(self.metric_metadata, None),
            start_time=(self.start_time, None),
            end_time=(self.end_time, None),
            usage_summary=(self.usage_summary, None),
            execution_status=(self.execution_status, None),
        ),
        exclude_default=True,
        **kwargs,
    )

  @classmethod
  def from_json(
      cls,
      json_value: dict[str, Any],
      *,
      example_input_by_id: Callable[[int], Any] | None = None,
      load_example_metadata: bool | Callable[['Example'], bool] = False,
      **kwargs
  ) -> 'Example':
    """Creates an example from the JSON representation."""
    example_id = json_value.get('id')
    if example_input_by_id:
      example_input = example_input_by_id(example_id)
    else:
      example_input = json_value.pop('input', pg.MISSING_VALUE)
      if example_input is not pg.MISSING_VALUE:
        example_input = pg.from_json(example_input, **kwargs)
    json_value['input'] = example_input

    # NOTE(daiyip): We need to load the types of the examples into the
    # deserialization context, otherwise the deserialization will fail if the
    # types are not registered.
    def example_class_defs(example) -> list[type[Any]]:
      referred_types = set()
      def _visit(k, v, p):
        del k, p
        if inspect.isclass(v):
          referred_types.add(v)
        elif isinstance(v, pg.Object):
          referred_types.add(v.__class__)
        return pg.TraverseAction.ENTER
      pg.traverse(example, _visit)
      return list(referred_types)

    # We delay loading the metadata until the other parts of the example are
    # loaded. So we could apply the filter to decide whether to load the
    # metadata.
    metadata_dict = json_value.pop('metadata', None)
    with pg.JSONConvertible.load_types_for_deserialization(
        *example_class_defs(example_input)
    ):
      example = cls(
          **{k: pg.from_json(v, **kwargs) for k, v in json_value.items()}
      )
      if callable(load_example_metadata):
        load_example_metadata = load_example_metadata(example)
      if load_example_metadata:
        example.metadata = pg.from_json(metadata_dict, **kwargs)
      return example

  @classmethod
  def iter_ckpts(
      cls,
      ckpt_file: str | list[str],
      example_input_by_id: Callable[[int], Any] | None = None,
      load_example_metadata: bool = True,
      convert_unknown: bool = True,
      **kwargs
  ) -> Iterator['Example']:
    """Iterates Examples from the checkpoint files."""
    ckpt_files = [ckpt_file] if isinstance(ckpt_file, str) else ckpt_file
    for ckpt_file in ckpt_files:
      with pg.io.open_sequence(ckpt_file) as f:
        for record in f:
          example = pg.from_json_str(
              record,
              example_input_by_id=example_input_by_id,
              load_example_metadata=load_example_metadata,
              convert_unknown=convert_unknown,
              **kwargs
          )
          assert isinstance(example, cls), example
          yield example

  #
  # HTML rendering.
  #

  def _html_tree_view_content(
      self,
      *,
      view: pg.views.HtmlTreeView,
      root_path: pg.KeyPath | None = None,
      extra_flags: dict[str, Any] | None = None,
      **kwargs
  ):
    root_path = root_path or pg.KeyPath()
    extra_flags = extra_flags or {}
    num_examples = extra_flags.get('num_examples', None)

    def _metric_label_group(metric_metadata: dict[str, Any] | None):
      """Renders a label group for metric metadata."""
      badges = []
      if metric_metadata:
        for metric_name, metadata in metric_metadata.items():
          assert isinstance(metadata, dict), (metric_name, metadata)
          for k, v in metadata.items():
            css_class = k
            if isinstance(v, bool):
              css_class += '_true' if v else '_false'
            badge = pg.views.html.controls.Badge(
                f'{k}:{v}',
                tooltip=f'{metric_name}: {k}',
                css_classes=[css_class],
            )
            badges.append(badge)
      return pg.views.html.controls.LabelGroup(badges)

    def _render_header():
      return pg.Html.element(
          'div',
          [
              pg.Html.element(
                  'div',
                  [
                      # Previous button.
                      pg.views.html.controls.Label(   # pylint: disable=g-long-ternary
                          '◀',
                          link=f'{self.id - 1}.html',
                          css_classes=['previous'],
                      ) if self.id > 1 else None,
                      # Current example ID.
                      pg.views.html.controls.Label(
                          f'#{self.id}',
                          css_classes=['example-id'],
                      ),
                      # Next button.
                      pg.views.html.controls.Label(   # pylint: disable=g-long-ternary
                          '▶',
                          link=f'{self.id + 1}.html',
                          css_classes=['next'],
                      ) if (num_examples is None
                            or self.id < num_examples) else None,

                  ]
              ),
              pg.Html.element(
                  'div',
                  [
                      # Usage summary.
                      pg.view(  # pylint: disable=g-long-ternary
                          self.usage_summary,
                          extra_flags=dict(as_badge=True)
                      ) if self.usage_summary is not None else None,
                      # Metric metadata.
                      _metric_label_group(self.metric_metadata)
                  ],
                  css_classes=['example-container'],
              )
          ]
      )

    def _render_content():
      def _tab(label, key, default):
        field = getattr(self, key)
        if default == field:
          return None
        return pg.views.html.controls.Tab(
            label=label,
            content=view.render(
                field,
                root_path=root_path + key,
                collapse_level=None,
                **view.get_passthrough_kwargs(**kwargs),
            ),
        )
      tabs = [
          _tab('Input', 'input', pg.MISSING_VALUE),
          _tab('Output', 'output', pg.MISSING_VALUE),
          _tab('Output Metadata', 'metadata', {}),
          _tab('Error', 'error', None),
      ]
      tabs = [tab for tab in tabs if tab is not None]
      return pg.views.html.controls.TabControl(
          tabs,
          len(tabs) - 1,
      )

    return pg.Html.element(
        'div',
        [
            _render_header(),
            _render_content(),
        ],
        css_classes=['eval-example']
    )

  def _html_tree_view_summary(self, *, view, **kwargs):
    return None

  @classmethod
  def _html_tree_view_css_styles(cls) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        .example-container {
          display: block;
          padding: 10px;
        }
        .example-id {
          font-weight: bold;
          font-size: 40px;
          margin: 0 10px;
          vertical-align: middle;
        }
        a.previous, a.next {
          text-decoration: none;
          vertical-align: middle;
          display: inline-block;
          padding: 8px 8px;
          color: #DDD;
        }
        a.previous:hover, a.next:hover {
          background-color: #ddd;
          color: black;
        }
        /* Badge styles. */
        .eval-example .badge.is_correct_true {
          color: green;
          background-color: #dcefbe;
        }
        .eval-example .badge.is_correct_false {
          color: orange;
          background-color: #ffefc4;
        }
        .eval-example .badge.error {
          color: red;
          background-color: #fdcccc;
        }
        .eval-example .badge.score {
          color: blue;
          background-color: #c4dced;
        }
        """
    ]




# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import inspect
import typing
import unittest

import langfun.core as lf
from langfun.core.llms import fake
from langfun.core.structured.schema import base
from langfun.core.structured.schema import python
import pyglove as pg


class Activity(pg.Object):
  description: str


class Itinerary(pg.Object):
  """A travel itinerary for a day."""

  day: pg.typing.Int[1, None]
  type: pg.typing.Enum['daytime', 'nighttime']
  activities: list[Activity]
  hotel: pg.typing.Annotated[
      pg.typing.Str['.*Hotel'] | None,
      'Hotel to stay if applicable.'
  ]


class PlaceOfInterest(pg.Object):
  """The name of a place of interest.

  A place of interest is a place that people ususally visit during their
  travels.
  """

  name: str


Itinerary.__serialization_key__ = 'Itinerary'


class Node(pg.Object):
  children: list['Node']


class SchemaReprTest(unittest.TestCase):

  def test_class_definition(self):
    self.assertEqual(
        python.class_definition(Activity, allowed_dependencies=set()),
        'class Activity:\n  description: str\n',
    )
    self.assertEqual(
        python.class_definition(Itinerary),
        inspect.cleandoc("""
            class Itinerary(Object):
              \"\"\"A travel itinerary for a day.\"\"\"
              day: int(min=1)
              type: Literal['daytime', 'nighttime']
              activities: list[Activity]
              # Hotel to stay if applicable.
              hotel: str(regex='.*Hotel') | None
            """) + '\n',
    )
    self.assertEqual(
        python.class_definition(
            PlaceOfInterest, allowed_dependencies=set()
        ),
        inspect.cleandoc("""
            class PlaceOfInterest:
              \"\"\"The name of a place of interest.

              A place of interest is a place that people ususally visit during their
              travels.
              \"\"\"
              name: str
            """) + '\n',
    )

    class A(pg.Object):
      pass

    self.assertEqual(
        python.class_definition(A, allowed_dependencies=set()),
        'class A:\n  pass\n',
    )
    self.assertEqual(
        python.class_definition(A),
        'class A(Object):\n  pass\n',
    )

    class C(pg.Object):
      x: str
      __kwargs__: typing.Any

    self.assertEqual(
        python.class_definition(C), 'class C(Object):\n  x: str\n'
    )

    class D(pg.Object):
      x: str
      @python.include_method_in_prompt
      def __call__(self, y: int) -> int:
        return len(self.x) + y

    self.assertEqual(
        python.class_definition(D),
        inspect.cleandoc(
            """
            class D(Object):
              x: str

              def __call__(self, y: int) -> int:
                return len(self.x) + y
            """) + '\n'
    )

    class E(pg.Object):
      x: str
      y: typing.Annotated[int, 'y', dict(exclude_from_prompt=True)]

    self.assertEqual(
        python.class_definition(E),
        inspect.cleandoc(
            """
            class E(Object):
              x: str
            """) + '\n'
    )

  def test_repr(self):
    class Foo(pg.Object):
      x: int

    @dataclasses.dataclass(frozen=True)
    class Bar:
      """Class Bar."""
      y: str

    @dataclasses.dataclass(frozen=True)
    class Baz(Bar):  # pylint: disable=unused-variable
      pass

    class A(pg.Object):
      foo: Foo

      @python.include_method_in_prompt
      def foo_value(self) -> int:
        return self.foo.x

      def baz_value(self) -> str:
        return 'baz'

    class B(A):
      bar: Bar
      foo2: Foo

      @python.include_method_in_prompt
      def bar_value(self) -> str:
        return self.bar.y

    schema = base.Schema([B])
    self.assertEqual(
        python.PythonPromptingProtocol().class_definitions(schema),
        inspect.cleandoc('''
            class Foo:
              x: int

            class Bar:
              """Class Bar."""
              y: str

            class Baz(Bar):
              """Baz(y: str)"""
              y: str

            class B:
              foo: Foo
              bar: Bar
              foo2: Foo

              def bar_value(self) -> str:
                return self.bar.y

              def foo_value(self) -> int:
                return self.foo.x
            ''') + '\n',
    )

    self.assertEqual(
        python.PythonPromptingProtocol().result_definition(schema), 'list[B]'
    )

    self.assertEqual(
        base.schema_repr(schema),
        inspect.cleandoc('''
            list[B]

            ```python
            class Foo:
              x: int

            class Bar:
              """Class Bar."""
              y: str

            class Baz(Bar):
              """Baz(y: str)"""
              y: str

            class B:
              foo: Foo
              bar: Bar
              foo2: Foo

              def bar_value(self) -> str:
                return self.bar.y

              def foo_value(self) -> int:
                return self.foo.x
            ```
            '''),
    )
    self.assertEqual(
        python.PythonPromptingProtocol().schema_repr(
            schema,
            include_result_definition=False,
            markdown=False,
        ),
        inspect.cleandoc('''
            class Foo:
              x: int

            class Bar:
              """Class Bar."""
              y: str

            class Baz(Bar):
              """Baz(y: str)"""
              y: str

            class B:
              foo: Foo
              bar: Bar
              foo2: Foo

              def bar_value(self) -> str:
                return self.bar.y

              def foo_value(self) -> int:
                return self.foo.x
            '''),
    )


class ValuePythonReprTest(unittest.TestCase):

  def test_repr(self):
    class Foo(pg.Object):
      x: int

    class A(pg.Object):
      foo: list[Foo]
      y: str | None

    self.assertEqual(
        base.value_repr(1, base.Schema(int)),
        '```python\n1\n```'
    )
    self.assertEqual(
        base.value_repr(lf.Template('hi, {{a}}', a='foo')),
        'hi, foo'
    )
    self.assertEqual(
        base.value_repr(
            A([Foo(1), Foo(2)], 'bar'), base.Schema(A), markdown=False,
        ),
        "A(foo=[Foo(x=1), Foo(x=2)], y='bar')",
    )
    self.assertEqual(
        base.value_repr(
            A([Foo(1), Foo(2)], 'bar'),
            base.Schema(A),
            markdown=True,
            compact=False,
            assign_to_var='output',
        ),
        inspect.cleandoc("""
            ```python
            output = A(
              foo=[
                Foo(
                  x=1
                ),
                Foo(
                  x=2
                )
              ],
              y='bar'
            )
            ```
            """),
    )
    self.assertEqual(
        base.value_repr(A),
        inspect.cleandoc("""
            ```python
            class Foo(Object):
              x: int

            class A(Object):
              foo: list[Foo]
              y: str | None
            ```
            """),
    )
    self.assertEqual(python.source_form(int), 'int')

  def test_parse(self):
    class Foo(pg.Object):
      x: int

    class A(pg.Object):
      foo: list[Foo]
      y: str | None

    self.assertEqual(
        base.parse_value(
            "A(foo=[Foo(x=1), Foo(x=2)], y='bar')", base.Schema(A)
        ),
        A([Foo(1), Foo(2)], y='bar'),
    )

  def test_parse_with_correction(self):
    class Foo(pg.Object):
      x: int

    class A(pg.Object):
      foo: list[Foo]
      y: str | None

    self.assertEqual(
        base.parse_value(
            "A(foo=[Foo(x=1), Foo(x=2)], y='bar'",
            base.Schema(A),
            autofix=1,
            autofix_lm=fake.StaticResponse(
                inspect.cleandoc(
                    """
                    CorrectedCode(
                        corrected_code='A(foo=[Foo(x=1), Foo(x=2)], y=\\\'bar\\\')',
                    )
                    """
                )
            ),
        ),
        A([Foo(1), Foo(2)], y='bar'),
    )

  def test_parse_class_def(self):
    self.assertTrue(
        inspect.isclass(
            base.parse_value(
                """
                class A:
                  x: Dict[str, Any]
                  y: Optional[Sequence[str]]
                  z: Union[int, List[int], Tuple[int]]
                """,
                permission=pg.coding.CodePermission.ALL,
            )
        )
    )


class StructureFromPythonTest(unittest.TestCase):

  def test_parse_class_def(self):

    class B:
      pass

    schema = base.Schema([B])
    v = python.structure_from_python(
        """
        class C(B):
          pass
        """,
        global_vars=dict(B=B),
        permission=pg.coding.CodePermission.ALL,
    )
    self.assertEqual(v.__module__, 'builtins')
    self.assertEqual(schema.class_dependencies(), [B])


if __name__ == '__main__':
  unittest.main()
